class Pdf2Text:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import PdfToText
        return PdfToText() \
            .setInputCol("content") \
            .setOutputCol("text") \
            .setPageNumCol("pagenum")




"""
Resolve Annotator Classes in the Pipeline to Extractor Configs and Methods.
Each Spark NLP Annotator Class is mapped to at least one

Every Annotator should have 2 configs. Some might offor multuple configs/method pairs, based on model_anno_obj/NLP reference.
- default/minimalistic -> Just the results of the annotations, no confidences or extra metadata
- with meta            -> A config that leverages white/black list and gets the most relevant metadata
- with positions       -> With Begins/Ends


"""
from sparknlp.annotator import *
from sparknlp.base import *
from nlu.pipe.extractors.extractor_configs_OS import *

OS_anno2config = {
    NerConverter: {
        'default': default_ner_converter_config,
        'default_full': default_full_config,
    },
    MultiClassifierDLModel: {
        'default': default_multi_classifier_dl_config,
        'default_full': default_full_config,

    },
    PerceptronModel: {
        'default': default_POS_config,
        'default_full': default_full_config,
    },
    ClassifierDLModel: {
        'default': default_classifier_dl_config,
        'default_full': default_full_config,
    },
    BertEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    BertForSequenceClassification : {
        'default': default_classifier_dl_config,
        'default_full': default_full_config,
    },
    DistilBertForSequenceClassification : {
        'default': default_classifier_dl_config,
        'default_full': default_full_config,
    },
    BertForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    DistilBertForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    RoBertaForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    XlmRoBertaForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    XlnetForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    AlbertForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    LongformerForTokenClassification: {
        'default': default_token_classifier_config,
        'default_full': default_full_config,
    },
    LongformerEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    AlbertEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    XlnetEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    RoBertaEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    XlmRoBertaEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    DistilBertEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    WordEmbeddingsModel: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    ElmoEmbeddings: {
        'default': default_word_embedding_config,
        'default_full': default_full_config,
    },
    BertSentenceEmbeddings: {
        'default': default_sentence_embedding_config,
        'default_full': default_full_config,
    },
    XlmRoBertaSentenceEmbeddings: {
        'default': default_sentence_embedding_config,
        'default_full': default_full_config,
    },
    Doc2VecModel: {
        'default': default_sentence_embedding_config,
        'default_full': default_full_config,
    },
    Doc2VecApproach: {
        'default': default_sentence_embedding_config,
        'default_full': default_full_config,
    },
    UniversalSentenceEncoder: {
        'default': default_sentence_embedding_config,
        'default_full': default_full_config,
    },
    SentenceEmbeddings: {
        'default': default_sentence_embedding_config,
        'default_full': default_full_config,
    },
    Tokenizer: {
        'default': default_tokenizer_config,
        'default_full': default_full_config,
    },
    TokenizerModel: {
        'default': default_tokenizer_config,
        'default_full': default_full_config,
    },
    RegexTokenizer: {
        'default': default_tokenizer_config,
        'default_full': default_full_config,
    },
    DocumentAssembler: {
        'default': default_document_config,
        'default_full': default_full_config,
    },
    SentenceDetectorDLModel: {
        'default': default_sentence_detector_DL_config,
        'default_full': default_full_config,
    },
    SentenceDetector: {
        'default': default_sentence_detector_config,
        'default_full': default_full_config,
    },
    ContextSpellCheckerModel: {
        'default': default_spell_context_config,
        'default_full': default_full_config,
    },
    SymmetricDeleteModel: {
        'default': default_spell_symmetric_config,
        'default_full': default_full_config,
    },
    NorvigSweetingModel: {
        'default': default_spell_norvig_config,
        'default_full': default_full_config,
    },
    LemmatizerModel: {
        'default': default_lemma_config,
        'default_full': default_full_config,
    },
    Normalizer: {
        'default': default_norm_config,
        'default_full': default_full_config,
    },
    NormalizerModel: {
        'default': default_norm_config,
        'default_full': default_full_config,
    },
    DocumentNormalizer: {
        'default': default_norm_document_config,
        'default_full': default_full_config,
    },
    Stemmer: {
        'default': default_stemm_config,
        'default_full': default_full_config,
    },
    NerDLModel: {
        'default': default_NER_config,
        'meta': meta_NER_config,
        'default_full': default_full_config,
    },
    NerCrfModel: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    LanguageDetectorDL: {
        'default': default_lang_classifier_config,
        'default_full': default_full_config,
    },
    DependencyParserModel: {
        'default': default_dep_untyped_config,
        'default_full': default_full_config,
    },
    TypedDependencyParserModel: {
        'default': default_dep_typed_config,
        'default_full': default_full_config,
    },
    SentimentDLModel: {
        'default': default_sentiment_dl_config,
        'default_full': default_full_config,
    },
    SentimentDetectorModel: {
        'default': default_sentiment_config,
        'default_full': default_full_config,
    },
    ViveknSentimentModel: {
        'default': default_sentiment_vivk_config,
        'default_full': default_full_config,
    },
    Chunker: {
        'default': default_chunk_config,
        'default_full': default_full_config,
    },
    NGramGenerator: {
        'default': default_ngram_config,
        'default_full': default_full_config,
    },
    ChunkEmbeddings: {
        'default': default_chunk_embedding_config,
        'default_full': default_full_config,
    },
    StopWordsCleaner: {
        'default': default_stopwords_config,
        'default_full': default_full_config,
    },
    TextMatcherModel: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    TextMatcher: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    RegexMatcherModel: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    RegexMatcher: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    DateMatcher: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    MultiDateMatcher: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    Doc2Chunk: {
        'default': default_doc2chunk_config,
        'default_full': default_full_config,
    },
    Chunk2Doc: {
        'default': '',  # TODO
        'default_full': default_full_config,
    },
    T5Transformer: {
        'default': default_T5_config,
        'default_full': default_full_config,
    },
    MarianTransformer: {
        'default': default_marian_config,
        'default_full': default_full_config,
    },
    YakeKeywordExtraction: {
        'default': default_yake_config,
        'default_full': default_full_config,
    },
    WordSegmenterModel: {
        'default': default_word_segmenter_config,
        'default_full': default_full_config,
    },
    # approaches
    ViveknSentimentApproach: {'default': '', 'default_full': default_full_config, },
    SentimentDLApproach: {'default': default_sentiment_dl_config, 'default_full': default_full_config, },
    ClassifierDLApproach: {'default': default_classifier_dl_config, 'default_full': default_full_config, },
    MultiClassifierDLApproach: {'default': default_multi_classifier_dl_config, 'default_full': default_full_config, },
    NerDLApproach: {'default': default_NER_config, 'default_full': default_full_config, },
    PerceptronApproach: {'default': default_POS_config, 'default_full': default_full_config, },

    # PretrainedPipeline : {
    #     'default' : '',
    # }
}



"""Utils for making final output columns generated by pythonify nicer
Col naming schema follows
<type>.<nlu_ref_identifier>.<field>
IF there is only 1 component_to_resolve of <type> in the component_list, the <type> will/can be ommitted.

- we remove all _<field> suffixex
- replace all '@' with '_'
"""
import logging
from typing import List

from nlu.universe.feature_node_ids import NLP_NODE_IDS
from sparknlp.annotator import *

import nlu
from nlu.pipe.col_substitution import substitution_map_OS
from nlu.pipe.extractors.extractor_base_data_classes import SparkOCRExtractorConfig
from nlu.universe.feature_universes import NLP_FEATURES
from nlu.universe.logic_universes import AnnoTypes
from nlu.universe.universes import Licenses

logger = logging.getLogger('nlu')

""" NAMING SCHEMAS after pythonify procedure : 
### NAMING RESULT SCHEMA: 


results         = { configs.output_col_prefix+'_results'  : list(map(unpack_result,row))} if configs.get_result else {}
beginnings      = { configs.output_col_prefix+'_beginnings' : list(map(unpack_begin,row))} if configs.get_begin or configs.get_positions else {}
endings         = { configs.output_col_prefix+'_endings'    : next(map(unpack_end,row))} if configs.get_end or configs.get_positions else {}
embeddings      = { configs.output_col_prefix+'_embeddings' : next(map(unpack_embeddings,row))} if configs.get_embeds else {}


### METADATA NAMING SCHEMA

result = dict(zip(list(map(lambda x : 'meta_'+ configs.output_col_prefix + '_' + x, keys_in_metadata)),meta_values_list))




"""
from sparknlp.base import MultiDocumentAssembler


class ColSubstitutionUtils:
    """Utils for substituting col names in Pythonify to short and meaningful names.
    Uses custom rename methods for either PySpark or Pandas
    """
    cleanable_splits = ['ner_converter', 'spell', 'ner_to_chunk_converter', 'train', 'classify', 'ner', 'med_ner', 'dl',
                        'match', 'clean', 'sentiment', 'embed', 'embed_sentence', 'embed_chunk', 'explain', 'pos',
                        'resolve_chunk', 'resolve', ]

    @staticmethod
    def substitute_col_names(df, anno_2_ex, pipe, stranger_cols=[], get_embeddings=False, drop_debug_cols=True):
        """
        Some truly irrelevant cols might be dropped, regardless of anno Extractor config
        Some truly irrelevant cols might be dropped, regardless of anno Extractor config
        0. Get list of annotator classes that are duplicates. Check inside the NLU Component Embelishment
        1. Get list of cols derived by component_to_resolve
        2. Substitute list of cols in DF with custom logic
        """
        substitution_fn = 'TODO'

        anno2final_cols = {}  # mapping of final col names to annotator class Key=AnnoModel, Value=List of Result cols
        new_cols = {}
        if pipe.has_licensed_components:
            from nlu.pipe.col_substitution import substitution_map_HC
        deducted_component_names = ColSubstitutionUtils.deduct_component_names(pipe)
        for c in pipe.components:
            if c.license == Licenses.ocr:
                # TODO better substitution
                old2new_anno_cols = {k: k for k in c.spark_output_column_names}
                anno2final_cols[c.model] = list(old2new_anno_cols.values())
                new_cols.update(old2new_anno_cols)
                new_cols = {**new_cols, **(old2new_anno_cols)}
                continue
            if 'embedding' in c.type and get_embeddings == False: continue
            cols_to_substitute = ColSubstitutionUtils.get_final_output_cols_of_component(c, df, anno_2_ex)
            if len(cols_to_substitute) == 0:
                # finisher cleaned components cols
                continue
            if type(c.model) in substitution_map_OS.OS_anno2substitution_fn.keys():
                substitution_fn = substitution_map_OS.OS_anno2substitution_fn[type(c.model)]['default']
            else:
                substitution_fn = c.pdf_col_name_substitutor

            if pipe.has_licensed_components and substitution_fn != 'TODO':
                if type(c.model) in substitution_map_HC.HC_anno2substitution_fn.keys():
                    substitution_fn = substitution_map_HC.HC_anno2substitution_fn[type(c.model)]['default']
            if substitution_fn == 'TODO':
                logger.info(f"Could not find substitution function for os_components={c}, leaving col names untouched")
                old2new_anno_cols = dict(zip(cols_to_substitute, cols_to_substitute))
                anno2final_cols[c.model] = list(old2new_anno_cols.values())
                new_cols.update(old2new_anno_cols)
                continue
            # dic, key=old_col, value=new_col. Some cols may be omitted and missing from the dic which are deemed irrelevant. Behaivour can be disabled by setting drop_debug_cols=False
            old2new_anno_cols = substitution_fn(c, cols_to_substitute, deducted_component_names[c])
            anno2final_cols[c.model] = list(old2new_anno_cols.values())
            new_cols = {**new_cols, **(old2new_anno_cols)}

        pipe.anno2final_cols = anno2final_cols
        cols_to_rename = list(new_cols.keys())
        for k in cols_to_rename:
            # some cols might not exist because no annotations generated, so we need to double check it really exists
            if k not in df.columns: del new_cols[k]
        return df.rename(columns=new_cols)[
            list(set(new_cols.values()).union(set(stranger_cols)))] if drop_debug_cols else \
            df.rename(columns=new_cols)

    @staticmethod
    def get_final_output_cols_of_component(c, df, anno_2_ex) -> List[str]:
        # get_final_output_cols_of_component(self.components[1], pretty_df, anno_2_ex_config)
        """Get's a list of all columns that have been derived in the pythonify procedure from the component_to_resolve
        os_components in dataframe df for anno_2_ex configs """
        og_output_col = c.spark_output_column_names[0]

        # may be missing because finisher cleaning
        if og_output_col not in anno_2_ex: return []
        configs = anno_2_ex[og_output_col]

        if c.name == NLP_NODE_IDS.FINISHER:
            result_cols = c.model.getOutputCols()
            if c.model.getIncludeMetadata():
                result_cols = result_cols + [f'{col}_metadata' for col in result_cols]
            return result_cols
        result_cols = []
        if isinstance(configs, SparkOCRExtractorConfig):
            # TODO better OCR-EX handling --> Col Name generator function which we use everywhere for unified col naming !!!!!
            return ['text']
        if isinstance(c.model, MultiDocumentAssembler):
            return [f'{NLP_FEATURES.DOCUMENT_QUESTION}_results', f'{NLP_FEATURES.DOCUMENT_QUESTION_CONTEXT}_results']

        if configs.get_annotator_type: result_cols.append(configs.output_col_prefix + '_types')
        if configs.get_result: result_cols.append(configs.output_col_prefix + '_results')
        if configs.get_begin or configs.get_positions: result_cols.append(configs.output_col_prefix + '_beginnings')
        if configs.get_end or configs.get_positions: result_cols.append(configs.output_col_prefix + '_endings')
        if configs.get_embeds: result_cols.append(configs.output_col_prefix + '_embeddings')
        if configs.get_origin: result_cols.append(configs.output_col_prefix + '_origin')
        # find all metadata fields generated by component_to_resolve
        for col in df.columns:
            if 'meta_' + configs.output_col_prefix in col:
                base_meta_prefix = 'meta_' + configs.output_col_prefix
                meta_col_name = base_meta_prefix + col.split(base_meta_prefix)[-1]
                if meta_col_name in df.columns:
                    # special case for overlapping names with _
                    if col.split(base_meta_prefix)[-1].split('_')[1].isnumeric() and not \
                            c.spark_output_column_names[0].split('_')[-1].isnumeric(): continue
                    if col.split(base_meta_prefix)[-1].split('_')[1].isnumeric() and \
                            c.spark_output_column_names[0].split('_')[-1].isnumeric():
                        id1 = int(col.split(base_meta_prefix)[-1].split('_')[1])
                        id2 = int(c.spark_output_column_names.split('_')[-1])
                        if id1 != id2: continue
                    result_cols.append(meta_col_name)
                elif c.type == AnnoTypes.CHUNK_CLASSIFIER:
                    result_cols.append(col)
                else:
                    logger.info(f"Could not find meta col for os_components={c}, col={col}. Ommiting col..")
        return result_cols

    @staticmethod
    def deduct_component_names(pipe):
        """Deduct a meaningful name for Embeddings, classifiers, resolvesr, relation extractors, etc..
        Will return a dict that maps every Annotator Class to a String Name. If String_Name =='' that means, it can be omtited for naming and the
        unique_default name schema should be used,
        since that annotator is unique in the component_list
        """
        # Todo extract name deductable as NLU component attribute
        import nlu.pipe.col_substitution.name_deduction.name_deductable_annotators_OS as deductable_OS
        max_depth = 10
        result_names = {}
        for c in pipe.components:
            is_partially_ready = c.type == AnnoTypes.PARTIALLY_READY
            if is_partially_ready or c.loaded_from_pretrained_pipe:
                if hasattr(c.model, 'getOutputCol'):
                    result_names[c] = c.model.getOutputCol()
                elif hasattr(c.model, 'getOutputCols'):
                    result_names[c] = c.model.getOutputCols()[0]
                else:
                    result_names[c] = str(c.model)
                continue

            result_names[c] = 'UNIQUE'  # assuemd uniqe, if not updated in followign steps
            is_always_name_deductable_component = False
            hc_deducted = False
            if pipe.has_licensed_components:
                import nlu.pipe.col_substitution.name_deduction.name_deductable_annotators_HC as deductable_HC
                if type(c.model) not in deductable_HC.name_deductable_HC and type(
                        c.model) not in deductable_OS.name_deductable_OS:
                    continue
                else:
                    hc_deducted = True
                if type(c.model) in deductable_HC.always_name_deductable_HC: is_always_name_deductable_component = True

            if type(c.model) not in deductable_OS.name_deductable_OS and not hc_deducted and not is_partially_ready:
                continue
            if type(c.model) in deductable_OS.always_name_deductable_OS:
                is_always_name_deductable_component = True

            same_components = []
            for other_c in pipe.components:
                if c is other_c: continue
                if c.type == other_c.type: same_components.append(other_c)
            if len(same_components) or is_always_name_deductable_component:
                # make sure each name is unique among the components of same type
                # if is_partially_ready and c.loaded_from_pretrained_pipe:
                cur_depth = 1
                other_names = [ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(other_c) for other_c in
                               same_components]
                c_name = ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(c)
                while c_name in other_names and cur_depth < max_depth:
                    cur_depth += 1
                    other_names = [ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(other_c) for other_c in
                                   same_components]
                    c_name = ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(c, cur_depth)
                result_names[c] = c_name
            else:
                result_names[c] = 'UNIQUE'  # no name insertion required
        return result_names

    @staticmethod
    def deduct_name_from_nlu_ref_at_depth(c, depth=1):
        if isinstance(c.model, MarianTransformer): return c.nlu_ref.split('xx.')[-1].replace('marian.', '')
        splits = c.nlu_ref.split('.')
        # remove all name irrelevant splits
        while len(splits) > 1 and (splits[0] in nlu.Spellbook.pretrained_models_references.keys() or splits[
            0] in ColSubstitutionUtils.cleanable_splits): splits.pop(0)
        if len(splits) == 0:
            if isinstance(c.model, (NerDLModel, NerConverter)): return 'ner'
            return c.nlu_ref.replace("@", "_")
        elif splits[0] == 'sentiment' and len(splits) == 1:
            return 'UNIQUE'
        else:
            return '_'.join(splits[:depth])



"""
Collection of universes shared across all libraries (NLP/HC/OCR), which are collections of atoms
"""
from dataclasses import dataclass
from typing import List, Union

from nlu.universe.atoms import JslAnnoId, JslFeature, ExternalFeature
from nlu.universe.feature_node_ids import NLP_NODE_IDS, OCR_NODE_IDS, NLP_HC_NODE_IDS
from nlu.universe.feature_universes import NLP_FEATURES, OCR_FEATURES, NLP_HC_FEATURES


### ____ Pipeline Graph Representation Logik Building Blocks ______

@dataclass
class FeatureNode:  # or Mode Node?
    """Defines a Node in a ML Dependency Feature Graph.
    Anno= Node, In = In arrows, Out = out Arrows
    Each NLU Component will output one of these
    NODE = Anno Class
    INPUTS = Array of ML-Features
    OUTPUTS = Array of ML-Features

    Used to cast the pipeline dependency resolution algorithm into an abstract grpah
    """

    JSL_generator_anno_class: JslAnnoId  # JSL Annotator that can generate this Triplet. Could be from OCR/JSL-Internal/Spark-NLP
    ins: List[JslFeature]  # JSL Annotator that can generate this Triplet
    outs: List[JslFeature]  # JSL Annotator that can generate this Triplet


@dataclass
class NlpFeatureNode:  # or Mode Node? (FeatureNode)
    """A node representation for a Spark OCR Annotator
    Used to cast the pipeline dependency resolution algorithm into an abstract grpah
    """
    node: Union[JslAnnoId]  # JSL Annotator that can generate this Triplet. Could be from OCR/JSL-Internal/Spark-NLP
    ins: List[JslFeature]  # JSL Annotator that can generate this Triplet
    outs: List[JslFeature]  # JSL Annotator that can generate this Triplet


class NlpHcFeatureNode(FeatureNode): pass


class OcrFeatureNode(FeatureNode): pass


class EXTERNAL_NODES():
    """
    Start Node definitions for the NLU Pipeline Graph completion logic
    These are analogus to the various input types NLU may accept
    """
    RAW_TEXT = ExternalFeature('text')
    NON_WHITESPACED_TEXT = ExternalFeature('non_whitespaced_text')  # i.e. Chinese, Russian, etc..
    # TODO define how its derivable, i.e Accepted input types that can be converted to spark DF types
    # str_array = 'str_array'
    #
    # pandas_df = 'pandas_df'
    # pd_series = 'pandas_series'
    #
    # np_array = 'pandas_series'
    #
    # img_path = 'pandas_series'
    # file_path = 'file_path' # todo more granuar, i.e. by file type?


@dataclass
class NLP_FEATURE_NODES:  # or Mode Node?
    """All avaiable Feature nodes in Spark NLP
    Used to cast the pipeline dependency resolution algorithm into an abstract grpah
    """
    # High Level NLP Feature Nodes
    E = EXTERNAL_NODES
    A = NLP_NODE_IDS
    F = NLP_FEATURES
    nodes = {
        A.INSTRUCTOR_SENTENCE_EMBEDDINGS: NlpFeatureNode(A.INSTRUCTOR_SENTENCE_EMBEDDINGS, [F.DOCUMENT], [F.SENTENCE_EMBEDDINGS]),

        A.E5_SENTENCE_EMBEDDINGS: NlpFeatureNode(A.E5_SENTENCE_EMBEDDINGS, [F.DOCUMENT],[F.SENTENCE_EMBEDDINGS]),
        A.MPNET_SENTENCE_EMBEDDINGS: NlpFeatureNode(A.MPNET_SENTENCE_EMBEDDINGS, [F.DOCUMENT], [F.SENTENCE_EMBEDDINGS]),

        A.PARTIALLY_IMPLEMENTED: NlpFeatureNode(A.PARTIALLY_IMPLEMENTED, [F.UNKOWN], [F.UNKOWN]),

        A.COREF_SPAN_BERT: NlpFeatureNode(A.COREF_SPAN_BERT, [F.DOCUMENT, F.TOKEN], [F.COREF_TOKEN]),

        A.BIG_TEXT_MATCHER: NlpFeatureNode(A.BIG_TEXT_MATCHER, [F.DOCUMENT, F.TOKEN], [F.CHUNK]),
        A.CHUNK2DOC: NlpFeatureNode(A.CHUNK2DOC, [F.NAMED_ENTITY_CONVERTED], [F.DOCUMENT_FROM_CHUNK]),
        A.CHUNK_EMBEDDINGS_CONVERTER: NlpFeatureNode(A.CHUNK_EMBEDDINGS_CONVERTER, [F.CHUNK, F.WORD_EMBEDDINGS],
                                                     [F.CHUNK_EMBEDDINGS]),
        A.CHUNK_TOKENIZER: NlpFeatureNode(A.CHUNK_TOKENIZER, [F.CHUNK], [F.TOKEN_CHUNKED]),
        A.CHUNKER: NlpFeatureNode(A.CHUNKER, [F.DOCUMENT, F.POS], [F.CHUNK]),
        A.CLASSIFIER_DL: NlpFeatureNode(A.CLASSIFIER_DL, [F.SENTENCE_EMBEDDINGS], [F.CATEGORY]),
        A.TRAINABLE_CLASSIFIER_DL: NlpFeatureNode(A.CLASSIFIER_DL, [F.SENTENCE_EMBEDDINGS], [F.CATEGORY]),
        A.CONTEXT_SPELL_CHECKER: NlpFeatureNode(A.CONTEXT_SPELL_CHECKER, [F.TOKEN], [F.TOKEN_SPELL_CHECKED]),
        A.DATE_MATCHER: NlpFeatureNode(A.DATE_MATCHER, [F.DOCUMENT], [F.DATE]),
        A.UNTYPED_DEPENDENCY_PARSER: NlpFeatureNode(A.UNTYPED_DEPENDENCY_PARSER, [F.DOCUMENT, F.POS, F.TOKEN],
                                                    [F.UNLABLED_DEPENDENCY]),
        A.TYPED_DEPENDENCY_PARSER: NlpFeatureNode(A.TYPED_DEPENDENCY_PARSER, [F.TOKEN, F.POS, F.UNLABLED_DEPENDENCY],
                                                  [F.LABELED_DEPENDENCY]),
        A.DOC2CHUNK: NlpFeatureNode(A.DOC2CHUNK, [F.DOCUMENT], [F.DOCUMENT_FROM_CHUNK]),

        A.MULTI_DOCUMENT_ASSEMBLER: NlpFeatureNode(A.MULTI_DOCUMENT_ASSEMBLER, [F.RAW_QUESTION, F.RAW_QUESTION_CONTEXT],
                                                   [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT]),
        A.TABLE_ASSEMBLER: NlpFeatureNode(A.TABLE_ASSEMBLER, [F.DOCUMENT_QUESTION_CONTEXT], [F.ASSEMBLED_TABULAR_DATA]),
        A.TAPAS_FOR_QA: NlpFeatureNode(A.TAPAS_FOR_QA, [F.DOCUMENT_QUESTION, F.ASSEMBLED_TABULAR_DATA, ],
                                       [F.TABULAR_ANSWER]),
        A.ALBERT_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.ALBERT_FOR_QUESTION_ANSWERING,
                                                        [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                        [F.CLASSIFIED_SPAN]),

        A.CAMEMBERT_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.CAMEMBERT_FOR_QUESTION_ANSWERING,
                                                           [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                           [F.CLASSIFIED_SPAN]),

        A.BERT_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.BERT_FOR_QUESTION_ANSWERING,
                                                      [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                      [F.CLASSIFIED_SPAN]),
        A.DE_BERTA_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.DE_BERTA_FOR_QUESTION_ANSWERING,
                                                          [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                          [F.CLASSIFIED_SPAN]),
        A.DISTIL_BERT_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.DISTIL_BERT_FOR_QUESTION_ANSWERING,
                                                             [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                             [F.CLASSIFIED_SPAN]),
        A.LONGFORMER_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.LONGFORMER_FOR_QUESTION_ANSWERING,
                                                            [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                            [F.CLASSIFIED_SPAN]),
        A.ROBERTA_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.ROBERTA_FOR_QUESTION_ANSWERING,
                                                         [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                         [F.CLASSIFIED_SPAN]),
        A.XLM_ROBERTA_FOR_QUESTION_ANSWERING: NlpFeatureNode(A.XLM_ROBERTA_FOR_QUESTION_ANSWERING,
                                                             [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT],
                                                             [F.CLASSIFIED_SPAN]),

        A.DOCUMENT_ASSEMBLER: NlpFeatureNode(A.DOCUMENT_ASSEMBLER, [E.RAW_TEXT], [F.DOCUMENT]),
        A.AUDIO_ASSEMBLER: NlpFeatureNode(A.AUDIO_ASSEMBLER, [F.RAW_AUDIO], [F.AUDIO]),
        # A.WAV2VEC_FOR_CTC: NlpFeatureNode(A.PARTIALLY_IMPLEMENTED, [F.AUDIO], [F.RECOGNIZED_SPEECH_TEXT]),
        A.WAV2VEC_FOR_CTC: NlpFeatureNode(A.WAV2VEC_FOR_CTC, [F.AUDIO], [E.RAW_TEXT]),
        A.HUBERT_FOR_CTC: NlpFeatureNode(A.HUBERT_FOR_CTC, [F.AUDIO], [E.RAW_TEXT]),
        A.WHISPER_FOR_CTC: NlpFeatureNode(A.WHISPER_FOR_CTC, [F.AUDIO], [E.RAW_TEXT]),

        A.IMAGE_ASSEMBLER: NlpFeatureNode(A.IMAGE_ASSEMBLER, [F.SPARK_NLP_IMAGE, F.SPARK_NLP_FILE_PATH], [F.IMAGE]),
        A.DOCUMENT_NORMALIZER: NlpFeatureNode(A.DOCUMENT_NORMALIZER, [F.DOCUMENT], [F.DOCUMENT_GENERATED]),
        A.EMBEDDINGS_FINISHER: NlpFeatureNode(A.EMBEDDINGS_FINISHER, [F.ANY_EMBEDDINGS], [F.FINISHED_EMBEDDINGS]),
        # A.# ENTITY_RULER : NlpFeatureNode(A.ENTITY_RULER, [F.], [F.]) # TODO? ,
        A.FINISHER: NlpFeatureNode(A.FINISHER, [F.ANY], [F.ANY_FINISHED]),
        A.GRAPH_EXTRACTION: NlpFeatureNode(A.GRAPH_EXTRACTION, [F.DOCUMENT, F.TOKEN, F.NAMED_ENTITY_IOB], [F.NODE]),
        # A.# GRAPH_FINISHER : NlpFeatureNode(A.GRAPH_FINISHER, [F.], [F.]) ,
        A.LANGUAGE_DETECTOR_DL: NlpFeatureNode(A.LANGUAGE_DETECTOR_DL, [F.DOCUMENT], [F.LANGUAGE]),
        A.LEMMATIZER: NlpFeatureNode(A.LEMMATIZER, [F.TOKEN], [F.TOKEN_LEMATIZED]),
        A.MULTI_CLASSIFIER_DL: NlpFeatureNode(A.MULTI_CLASSIFIER_DL, [F.SENTENCE_EMBEDDINGS],
                                              [F.MULTI_DOCUMENT_CLASSIFICATION]),
        A.TRAINABLE_MULTI_CLASSIFIER_DL: NlpFeatureNode(A.MULTI_CLASSIFIER_DL, [F.SENTENCE_EMBEDDINGS],
                                                        [F.MULTI_DOCUMENT_CLASSIFICATION]),
        A.MULTI_DATE_MATCHER: NlpFeatureNode(A.MULTI_DATE_MATCHER, [F.DOCUMENT], [F.DATE]),
        A.N_GRAMM_GENERATOR: NlpFeatureNode(A.N_GRAMM_GENERATOR, [F.TOKEN], [F.CHUNK]),
        A.NER_CONVERTER: NlpFeatureNode(A.NER_CONVERTER, [F.TOKEN, F.DOCUMENT, F.NAMED_ENTITY_IOB],
                                        [F.NAMED_ENTITY_CONVERTED]),
        A.NER_CRF: NlpFeatureNode(A.NER_CRF, [F.DOCUMENT, F.TOKEN, F.WORD_EMBEDDINGS], [F.NAMED_ENTITY_IOB]),
        A.NER_DL: NlpFeatureNode(A.NER_DL, [F.DOCUMENT, F.TOKEN, F.WORD_EMBEDDINGS], [F.NAMED_ENTITY_IOB]),
        A.TRAINABLE_NER_DL: NlpFeatureNode(A.TRAINABLE_NER_DL, [F.DOCUMENT, F.TOKEN, F.WORD_EMBEDDINGS],
                                           [F.NAMED_ENTITY_IOB]),
        A.NER_OVERWRITER: NlpFeatureNode(A.NER_OVERWRITER, [F.NAMED_ENTITY_IOB], [F.NAMED_ENTITY_IOB]),
        A.NORMALIZER: NlpFeatureNode(A.NORMALIZER, [F.TOKEN], [F.TOKEN_NORMALIZED]),
        A.NORVIG_SPELL_CHECKER: NlpFeatureNode(A.NORVIG_SPELL_CHECKER, [F.TOKEN], [F.TOKEN_SPELL_CHECKED]),
        A.POS: NlpFeatureNode(A.POS, [F.TOKEN, F.DOCUMENT], [F.POS]),
        A.TRAINABLE_POS: NlpFeatureNode(A.POS, [F.TOKEN, F.DOCUMENT], [F.POS]),
        A.RECURISVE_TOKENIZER: NlpFeatureNode(A.RECURISVE_TOKENIZER, [F.DOCUMENT], [F.TOKEN]),
        A.REGEX_MATCHER: NlpFeatureNode(A.REGEX_MATCHER, [F.DOCUMENT], [F.NAMED_ENTITY_CONVERTED]),
        A.REGEX_TOKENIZER: NlpFeatureNode(A.REGEX_TOKENIZER, [F.DOCUMENT], [F.TOKEN]),
        A.SENTENCE_DETECTOR: NlpFeatureNode(A.SENTENCE_DETECTOR, [F.DOCUMENT], [F.SENTENCE]),
        A.SENTENCE_DETECTOR_DL: NlpFeatureNode(A.SENTENCE_DETECTOR_DL, [F.DOCUMENT], [F.SENTENCE]),
        A.SENTENCE_EMBEDDINGS_CONVERTER: NlpFeatureNode(A.SENTENCE_EMBEDDINGS_CONVERTER,
                                                        [F.DOCUMENT, F.WORD_EMBEDDINGS], [F.SENTENCE_EMBEDDINGS]),
        A.SENTIMENT_DL: NlpFeatureNode(A.SENTIMENT_DL, [F.SENTENCE_EMBEDDINGS], [F.DOCUMENT_CLASSIFICATION]),
        A.TRAINABLE_SENTIMENT_DL: NlpFeatureNode(A.TRAINABLE_SENTIMENT_DL, [F.SENTENCE_EMBEDDINGS],
                                                 [F.DOCUMENT_CLASSIFICATION]),

        # A.# SENTENCE_DETECTOR : NlpFeatureNode(A.SENTENCE_DETECTOR, [F.TOKEN, F.DOCUMENT], [F.DOCUMENT_CLASSIFICATION] ,
        A.STEMMER: NlpFeatureNode(A.STEMMER, [F.TOKEN], [F.TOKEN_STEMMED]),
        A.STOP_WORDS_CLEANER: NlpFeatureNode(A.STOP_WORDS_CLEANER, [F.TOKEN], [F.TOKEN_STOP_WORD_REMOVED]),
        A.SYMMETRIC_DELETE_SPELLCHECKER: NlpFeatureNode(A.SYMMETRIC_DELETE_SPELLCHECKER, [F.TOKEN],
                                                        [F.TOKEN_SPELL_CHECKED]),
        A.TEXT_MATCHER: NlpFeatureNode(A.TEXT_MATCHER, [F.DOCUMENT, F.TOKEN], [F.CHUNK]),
        A.TOKEN2CHUNK: NlpFeatureNode(A.TOKEN2CHUNK, [F.TOKEN], [F.CHUNK]),
        A.TOKEN_ASSEMBLER: NlpFeatureNode(A.TOKEN_ASSEMBLER, [F.DOCUMENT, F.TOKEN], [F.DOCUMENT]),
        A.TOKENIZER: NlpFeatureNode(A.TOKENIZER, [F.DOCUMENT], [F.TOKEN]),
        A.VIVEKN_SENTIMENT: NlpFeatureNode(A.VIVEKN_SENTIMENT, [F.TOKEN, F.DOCUMENT], [F.DOCUMENT_CLASSIFICATION]),
        A.SENTIMENT_DETECTOR: NlpFeatureNode(A.SENTIMENT_DETECTOR, [F.TOKEN, F.DOCUMENT], [F.DOCUMENT_CLASSIFICATION]),
        A.WORD_EMBEDDINGS: NlpFeatureNode(A.WORD_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.WORD_SEGMENTER: NlpFeatureNode(A.WORD_SEGMENTER, [F.DOCUMENT], [F.TOKEN]),
        A.YAKE_KEYWORD_EXTRACTION: NlpFeatureNode(A.YAKE_KEYWORD_EXTRACTION, [F.TOKEN], [F.CHUNK]),
        A.ALBERT_EMBEDDINGS: NlpFeatureNode(A.ALBERT_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),

        A.CAMEMBERT_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.CAMEMBERT_FOR_TOKEN_CLASSIFICATION,
                                                             [F.DOCUMENT, F.TOKEN],
                                                             [F.TOKEN_CLASSIFICATION]),
        A.DEBERTA_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.DEBERTA_FOR_TOKEN_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                           [F.TOKEN_CLASSIFICATION]),
        A.ALBERT_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.ALBERT_FOR_TOKEN_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                          [F.TOKEN_CLASSIFICATION]),
        A.BERT_EMBEDDINGS: NlpFeatureNode(A.BERT_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.CAMEMBERT_EMBEDDINGS: NlpFeatureNode(A.CAMEMBERT_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.DEBERTA_WORD_EMBEDDINGS: NlpFeatureNode(A.BERT_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.BERT_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.BERT_FOR_TOKEN_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                        [F.TOKEN_CLASSIFICATION]),
        A.BERT_SENTENCE_EMBEDDINGS: NlpFeatureNode(A.BERT_SENTENCE_EMBEDDINGS, [F.DOCUMENT], [F.SENTENCE_EMBEDDINGS]),
        A.DISTIL_BERT_EMBEDDINGS: NlpFeatureNode(A.DISTIL_BERT_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.DISTIL_BERT_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.DISTIL_BERT_FOR_TOKEN_CLASSIFICATION,
                                                               [F.DOCUMENT, F.TOKEN], [F.TOKEN_CLASSIFICATION]),
        A.ELMO_EMBEDDINGS: NlpFeatureNode(A.ELMO_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.LONGFORMER_EMBEDDINGS: NlpFeatureNode(A.LONGFORMER_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.LONGFORMER_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.LONGFORMER_FOR_TOKEN_CLASSIFICATION,
                                                              [F.DOCUMENT, F.TOKEN], [F.TOKEN_CLASSIFICATION]),
        A.MARIAN_TRANSFORMER: NlpFeatureNode(A.MARIAN_TRANSFORMER, [F.DOCUMENT], [F.DOCUMENT_TRANSLATED]),
        A.ROBERTA_EMBEDDINGS: NlpFeatureNode(A.ROBERTA_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.ROBERTA_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.ROBERTA_FOR_TOKEN_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                           [F.TOKEN_CLASSIFICATION]),
        A.ROBERTA_SENTENCE_EMBEDDINGS: NlpFeatureNode(A.ROBERTA_SENTENCE_EMBEDDINGS, [F.DOCUMENT], [F.SENTENCE_EMBEDDINGS]),

        A.T5_TRANSFORMER: NlpFeatureNode(A.T5_TRANSFORMER, [F.DOCUMENT], [F.DOCUMENT_GENERATED]),
        A.UNIVERSAL_SENTENCE_ENCODER: NlpFeatureNode(A.UNIVERSAL_SENTENCE_ENCODER, [F.DOCUMENT],
                                                     [F.SENTENCE_EMBEDDINGS]),
        A.XLM_ROBERTA_EMBEDDINGS: NlpFeatureNode(A.XLM_ROBERTA_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.XLM_ROBERTA_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.XLM_ROBERTA_FOR_TOKEN_CLASSIFICATION,
                                                               [F.DOCUMENT, F.TOKEN], [F.TOKEN_CLASSIFICATION]),
        A.XLM_ROBERTA_SENTENCE_EMBEDDINGS: NlpFeatureNode(A.XLM_ROBERTA_SENTENCE_EMBEDDINGS, [F.DOCUMENT],
                                                          [F.SENTENCE_EMBEDDINGS]),
        A.XLNET_EMBEDDINGS: NlpFeatureNode(A.XLNET_EMBEDDINGS, [F.DOCUMENT, F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.XLNET_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.XLNET_FOR_TOKEN_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                         [F.TOKEN_CLASSIFICATION]),

        A.DOC2VEC: NlpFeatureNode(A.DOC2VEC, [F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.TRAIANBLE_DOC2VEC: NlpFeatureNode(A.TRAIANBLE_DOC2VEC, [F.TOKEN], [F.WORD_EMBEDDINGS]),

        A.BERT_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.BERT_FOR_SEQUENCE_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                           [F.SEQUENCE_CLASSIFICATION]),

        A.CAMEMBERT_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.CAMEMBERT_FOR_SEQUENCE_CLASSIFICATION,
                                                                [F.DOCUMENT, F.TOKEN],
                                                                [F.SEQUENCE_CLASSIFICATION]),
        A.BERT_FOR_ZERO_SHOT_CLASSIFICATION: NlpFeatureNode(A.BERT_FOR_ZERO_SHOT_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                            [F.SEQUENCE_CLASSIFICATION]),

        A.DEBERTA_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.BERT_FOR_SEQUENCE_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
                                                              [F.SEQUENCE_CLASSIFICATION]),

        A.DISTIL_BERT_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.DISTIL_BERT_FOR_SEQUENCE_CLASSIFICATION,
                                                                  [F.DOCUMENT, F.TOKEN],
                                                                  [F.SEQUENCE_CLASSIFICATION]),
        A.DISTIL_BERT_FOR_ZERO_SHOT_CLASSIFICATION: NlpFeatureNode(A.DISTIL_BERT_FOR_ZERO_SHOT_CLASSIFICATION,
                                                                   [F.DOCUMENT, F.TOKEN],
                                                                   [F.SEQUENCE_CLASSIFICATION]),

        A.XLM_ROBERTA_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.XLM_ROBERTA_FOR_SEQUENCE_CLASSIFICATION,
                                                                  [F.DOCUMENT, F.TOKEN],
                                                                  [F.SEQUENCE_CLASSIFICATION]),
        A.ROBERTA_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.ROBERTA_FOR_SEQUENCE_CLASSIFICATION,
                                                              [F.DOCUMENT, F.TOKEN],
                                                              [F.SEQUENCE_CLASSIFICATION]),
        A.LONGFORMER_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.LONGFORMER_FOR_SEQUENCE_CLASSIFICATION,
                                                                 [F.DOCUMENT, F.TOKEN],
                                                                 [F.SEQUENCE_CLASSIFICATION]),
        A.ALBERT_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.ALBERT_FOR_SEQUENCE_CLASSIFICATION,
                                                             [F.DOCUMENT, F.TOKEN],
                                                             [F.SEQUENCE_CLASSIFICATION]),
        A.XLNET_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.XLNET_FOR_SEQUENCE_CLASSIFICATION,
                                                            [F.DOCUMENT, F.TOKEN],
                                                            [F.SEQUENCE_CLASSIFICATION]),
        A.ROBERTA_FOR_ZERO_SHOT_CLASSIFICATION: NlpFeatureNode(A.ROBERTA_FOR_ZERO_SHOT_CLASSIFICATION,
                                                               [F.DOCUMENT, F.TOKEN],
                                                               [F.SEQUENCE_CLASSIFICATION]),
        A.GPT2: NlpFeatureNode(A.GPT2, [F.DOCUMENT], [F.DOCUMENT_GENERATED]),
        A.WORD_2_VEC: NlpFeatureNode(A.WORD_2_VEC, [F.TOKEN], [F.WORD_EMBEDDINGS]),
        A.BERT_SENTENCE_CHUNK_EMBEDDINGS: NlpFeatureNode(A.BERT_SENTENCE_CHUNK_EMBEDDINGS, [F.DOCUMENT],
                                                         [F.NAMED_ENTITY_CONVERTED]),

        A.VIT_IMAGE_CLASSIFICATION: NlpFeatureNode(A.VIT_IMAGE_CLASSIFICATION, [F.IMAGE], [F.CLASSIFIED_IMAGE]),
        A.CONVNEXT_IMAGE_CLASSIFICATION: NlpFeatureNode(A.CONVNEXT_IMAGE_CLASSIFICATION, [F.IMAGE], [F.CLASSIFIED_IMAGE]),
        A.SWIN_IMAGE_CLASSIFICATION: NlpFeatureNode(A.SWIN_IMAGE_CLASSIFICATION, [F.IMAGE], [F.CLASSIFIED_IMAGE]),
        A.BART_TRANSFORMER: NlpFeatureNode(A.BART_TRANSFORMER, [F.DOCUMENT], [F.DOCUMENT_GENERATED]),

    }

@dataclass
class OCR_FEATURE_NODES:
    """All avaiable Feature nodes in OCR
    Used to cast the pipeline dependency resolution algorithm into an abstract grpah
    """

    # Visual Document UnderstandingBINARY2IMAGE
    A = OCR_NODE_IDS
    F = OCR_FEATURES
    nodes = {
        A.VISUAL_DOCUMENT_CLASSIFIER: OcrFeatureNode(A.VISUAL_DOCUMENT_CLASSIFIER, [F.HOCR],
                                                     [F.VISUAL_CLASSIFIER_PREDICTION, F.VISUAL_CLASSIFIER_CONFIDENCE]),

        A.IMAGE2HOCR: OcrFeatureNode(A.IMAGE2HOCR, [F.OCR_IMAGE], [F.HOCR]),

        # VISUAL_DOCUMENT_NER : OcrFeatureNode(A.VISUAL_DOCUMENT_NER, [OcrFeature.HOCR, OcrFeature.FILE_PATH], [NlpFeature.NER_Annotation]), # TODO NlpFeature Space!

        # Object Detection
        A.IMAGE_SPLIT_REGIONS: OcrFeatureNode(A.IMAGE_SPLIT_REGIONS, [F.OCR_IMAGE,F.OCR_REGION], [F.IMG_SPLIT_REGIONS]),

        # VISUAL_DOCUMENT_NER : OcrFeatureNode(A.VISUAL_DOCUMENT_NER, [OcrFeature.HOCR, OcrFeature.FILE_PATH], [NlpFeature.NER_Annotation]), # TODO NlpFeature Space!

        # Object Detection
        A.IMAGE_HANDWRITTEN_DETECTOR: OcrFeatureNode(A.IMAGE_HANDWRITTEN_DETECTOR, [F.OCR_IMAGE], [F.OCR_REGION]),

        # TABLE Processors/Recognition
        A.IMAGE_TABLE_DETECTOR: OcrFeatureNode(A.IMAGE_TABLE_DETECTOR, [F.OCR_IMAGE], [F.OCR_REGION]),
        A.IMAGE_TABLE_CELL_DETECTOR: OcrFeatureNode(A.IMAGE_TABLE_CELL_DETECTOR, [F.IMG_SPLIT_REGIONS, F.OCR_REGION],
                                                    [F.OCR_TABLE_CELLS]),
        A.IMAGE_TABLE_CELL2TEXT_TABLE: OcrFeatureNode(A.IMAGE_TABLE_CELL2TEXT_TABLE, [F.IMG_SPLIT_REGIONS,
                                                                                      F.OCR_TABLE_CELLS],
                                                      [F.OCR_TABLE]),
        # PDF Processing
        A.PDF2TEXT: OcrFeatureNode(A.PDF2TEXT, [F.BINARY_PDF, F.FILE_PATH], [F.TEXT, F.PAGE_NUM]),
        A.PDF2IMAGE: OcrFeatureNode(A.PDF2IMAGE, [F.BINARY_PDF, F.FILE_PATH, F.FALL_BACK], [F.OCR_IMAGE, F.PAGE_NUM]),
        A.IMAGE2PDF: OcrFeatureNode(A.IMAGE2PDF, [F.OCR_IMAGE, F.FILE_PATH], [F.BINARY_PDF]),
        A.TEXT2PDF: OcrFeatureNode(A.TEXT2PDF, [F.OCR_POSITIONS, F.OCR_IMAGE, F.OCR_TEXT, F.FILE_PATH, F.BINARY_PDF],
                                   [F.BINARY_PDF]),
        A.PDF_ASSEMBLER: OcrFeatureNode(A.PDF_ASSEMBLER, [F.BINARY_PDF_PAGE, F.FILE_PATH, F.PAGE_NUM], [F.BINARY_PDF]),
        A.PDF_DRAW_REGIONS: OcrFeatureNode(A.PDF_DRAW_REGIONS, [F.BINARY_PDF, F.FILE_PATH, F.OCR_POSITIONS],
                                           [F.BINARY_PDF]),
        A.PDF2TEXT_TABLE: OcrFeatureNode(A.PDF2TEXT_TABLE, [F.BINARY_DOCX, F.FILE_PATH, ], [F.OCR_TABLE]),

        # DOCX Processing
        A.DOC2TEXT: OcrFeatureNode(A.DOC2TEXT, [F.BINARY_DOCX, F.FILE_PATH, ], [F.TEXT, F.PAGE_NUM]),
        A.DOC2TEXT_TABLE: OcrFeatureNode(A.DOC2TEXT_TABLE, [F.BINARY_DOCX, F.FILE_PATH], [F.OCR_TABLE]),
        A.DOC2PDF: OcrFeatureNode(A.DOC2PDF, [F.BINARY_DOCX, F.FILE_PATH], [F.BINARY_PDF]),
        A.PPT2TEXT_TABLE: OcrFeatureNode(A.PPT2TEXT_TABLE, [F.BINARY_DOCX, F.FILE_PATH], [F.OCR_TABLE]),
        A.PPT2PDF: OcrFeatureNode(A.PPT2PDF, [F.BINARY_PPT, F.FILE_PATH], [F.BINARY_PDF]),

        # DICOM Processing
        A.DICOM2IMAGE: OcrFeatureNode(A.DICOM2IMAGE, [F.BINARY_DICOM, F.FILE_PATH],
                                      [F.OCR_IMAGE, F.PAGE_NUM, F.DICOM_METADATA]),
        A.IMAGE2DICOM: OcrFeatureNode(A.IMAGE2DICOM, [F.OCR_IMAGE, F.FILE_PATH, F.DICOM_METADATA], [F.BINARY_DICOM]),
        # Image Pre-Processing
        A.BINARY2IMAGE: OcrFeatureNode(A.BINARY2IMAGE, [F.BINARY_IMG, F.FILE_PATH], [F.OCR_IMAGE]),
        A.GPU_IMAGE_TRANSFORMER: OcrFeatureNode(A.GPU_IMAGE_TRANSFORMER, [F.OCR_IMAGE], [F.OCR_IMAGE]),

        A.IMAGE_BINARIZER: OcrFeatureNode(A.IMAGE_BINARIZER, [F.OCR_IMAGE], [F.OCR_IMAGE]),
        A.IMAGE_ADAPTIVE_BINARIZER: OcrFeatureNode(A.IMAGE_ADAPTIVE_BINARIZER, [F.OCR_IMAGE], [F.OCR_IMAGE]),
        A.IMAGE_ADAPTIVE_THRESHOLDING: OcrFeatureNode(A.IMAGE_ADAPTIVE_THRESHOLDING, [F.OCR_IMAGE], [F.OCR_IMAGE]),
        A.IMAGE_SCALER: OcrFeatureNode(A.IMAGE_SCALER, [F.OCR_IMAGE], [F.OCR_IMAGE]),
        A.IMAGE_ADAPTIVE_SCALER: OcrFeatureNode(A.IMAGE_ADAPTIVE_SCALER, [F.OCR_IMAGE], [F.OCR_IMAGE]),
        A.IMAGE_SKEW_CORRECTOR: OcrFeatureNode(A.IMAGE_SKEW_CORRECTOR, [F.OCR_IMAGE], [F.OCR_IMAGE]),

        # TODO THESE ALL BLOW??? Region???
        A.IMAGE_NOISE_SCORER: OcrFeatureNode(A.IMAGE_NOISE_SCORER, [F.OCR_IMAGE, F.OCR_REGION], [F.OCR_IMAGE]),
        # TODO WHAT IS REGION???? There is no schema for that
        A.IMAGE_REMOVE_OBJECTS: OcrFeatureNode(A.IMAGE_REMOVE_OBJECTS, [F.OCR_IMAGE], [F.OCR_IMAGE]),  # TODO
        A.IMAGE_MORPHOLOGY_OPERATION: OcrFeatureNode(A.IMAGE_MORPHOLOGY_OPERATION, [F.OCR_IMAGE], [F.OCR_IMAGE]),
        # TODO
        A.IMAGE_CROPPER: OcrFeatureNode(A.IMAGE_CROPPER, [F.OCR_IMAGE], [F.OCR_IMAGE]),  # TODO
        A.IMAGE2REGION: OcrFeatureNode(A.IMAGE2PDF, [F.OCR_IMAGE], [F.OCR_IMAGE]),  # TODO
        A.IMAGE_LAYOUT_ANALZYER: OcrFeatureNode(A.IMAGE_LAYOUT_ANALZYER, [F.OCR_IMAGE], [F.OCR_IMAGE]),  # TODO
        # A.IMAGE_SPLIT_REGIONS: OcrFeatureNode(A.IMAGE_SPLIT_REGIONS, [F.OCR_IMAGE], [F.OCR_IMAGE]),  # TODO
        A.IMAGE_DRAW_REGIONS: OcrFeatureNode(A.IMAGE_DRAW_REGIONS, [F.OCR_IMAGE], [F.OCR_IMAGE]),  # TODO

        # Character Recognition .. TODO these should be correct but not 100% sure about the positions
        A.IMAGE2TEXT: OcrFeatureNode(A.IMAGE2TEXT, [F.OCR_IMAGE], [F.TEXT, F.OCR_POSITIONS]),
        A.IMAGE2TEXTPDF: OcrFeatureNode(A.IMAGE2TEXTPDF, [F.OCR_IMAGE, F.FILE_PATH, F.PAGE_NUM], [F.BINARY_PDF]),

        # TODO is ouput HOCR format as in HOCR_DOCUMENT_ASSAMBLER???
        A.IMAGE_BRANDS2TEXT: OcrFeatureNode(A.IMAGE_BRANDS2TEXT, [F.OCR_IMAGE], [F.OCR_POSITIONS, F.TEXT,
                                                                                 F.OCR_IMAGE]),
        # TODO what is the STRUCTURE of output image_brand ??? OCR_IE??
        A.POSITION_FINDER: OcrFeatureNode(A.POSITION_FINDER, [F.TEXT_ENTITY, F.OCR_PAGE_MATRIX], [F.OCR_POSITIONS]),
        # TODO COORDINATE::POSITION??
        ##  TODO Updates text at a position? I.e. Change the text at given corodinates BUT THEN why is output position???
        A.UPDATE_TEXT_POSITION: OcrFeatureNode(A.POSITION_FINDER, [F.OCR_POSITIONS, F.TEXT_ENTITY], [F.OCR_POSITIONS]),
        # TODO COORDINATE::POSITION??
        ## Cancer Document Test parser. Required Text of Header Field of something
        A.FOUNDATION_ONE_REPORT_PARSER: OcrFeatureNode(A.FOUNDATION_ONE_REPORT_PARSER, [F.OCR_TEXT, F.FILE_PATH],
                                                       [F.JSON_FOUNDATION_ONE_REPORT]),
        # HOCR
        A.HOCR_DOCUMENT_ASSEMBLER: OcrFeatureNode(A.HOCR_TOKENIZER, [F.HOCR], [F.TEXT_DOCUMENT]),

        A.HOCR_TOKENIZER: OcrFeatureNode(A.HOCR_TOKENIZER, [F.HOCR], [F.TEXT_DOCUMENT_TOKENIZED]),
    }


@dataclass
class NLP_HC_FEATURE_NODES():
    """All avaiable Feature nodes in NLP Healthcare
    Used to cast the pipeline dependency resolution algorithm into an abstract grpah

    """
    # Visual Document Understanding
    A = NLP_HC_NODE_IDS
    F = NLP_FEATURES
    H_F = NLP_HC_FEATURES
    # HC Feature Nodes
    nodes = {
        A.MEDICAL_QUESTION_ANSWERING: NlpFeatureNode(A.MEDICAL_QUESTION_ANSWERING, [F.DOCUMENT_QUESTION, F.DOCUMENT_QUESTION_CONTEXT], [F.CLASSIFIED_SPAN]),

        A.MEDICAL_TEXT_GENERATOR: NlpFeatureNode(A.MEDICAL_TEXT_GENERATOR, [F.DOCUMENT], [F.DOCUMENT_GENERATED]),

        A.MEDICAL_SUMMARIZER: NlpFeatureNode(A.MEDICAL_SUMMARIZER, [F.DOCUMENT], [F.DOCUMENT_GENERATED]),

        A.ZERO_SHOT_NER: NlpHcFeatureNode(A.ZERO_SHOT_NER, [F.TOKEN, F.DOCUMENT],
                                          [F.NAMED_ENTITY_IOB]),
        A.CHUNK_MAPPER_MODEL: NlpHcFeatureNode(A.CHUNK_MAPPER_MODEL, [F.NAMED_ENTITY_CONVERTED],
                                               [H_F.MAPPED_CHUNK]),

        A.ASSERTION_DL: NlpHcFeatureNode(A.ASSERTION_DL, [F.DOCUMENT, F.NAMED_ENTITY_CONVERTED, F.WORD_EMBEDDINGS],
                                         [H_F.ASSERTION]),
        A.TRAINABLE_ASSERTION_DL: NlpHcFeatureNode(A.TRAINABLE_ASSERTION_DL,
                                                   [F.DOCUMENT, F.NAMED_ENTITY_CONVERTED, F.WORD_EMBEDDINGS],
                                                   [H_F.ASSERTION]),
        A.ASSERTION_FILTERER: NlpHcFeatureNode(A.ASSERTION_FILTERER, [F.DOCUMENT, F.CHUNK, H_F.ASSERTION], [F.CHUNK]),
        A.ASSERTION_LOG_REG: NlpHcFeatureNode(A.ASSERTION_LOG_REG, [F.DOCUMENT, F.CHUNK, F.WORD_EMBEDDINGS],
                                              [H_F.ASSERTION]),
        A.TRAINABLE_ASSERTION_LOG_REG: NlpHcFeatureNode(A.TRAINABLE_ASSERTION_LOG_REG,
                                                        [F.DOCUMENT, F.CHUNK, F.WORD_EMBEDDINGS], [H_F.ASSERTION]),
        A.CHUNK2TOKEN: NlpHcFeatureNode(A.CHUNK2TOKEN, [F.CHUNK], [F.TOKEN]),
        A.CHUNK_ENTITY_RESOLVER: NlpHcFeatureNode(A.CHUNK_ENTITY_RESOLVER, [F.TOKEN, F.WORD_EMBEDDINGS],
                                                  [H_F.RESOLVED_ENTITY]),
        A.TRAINABLE_CHUNK_ENTITY_RESOLVER: NlpHcFeatureNode(A.TRAINABLE_CHUNK_ENTITY_RESOLVER,
                                                            [F.TOKEN, F.WORD_EMBEDDINGS], [H_F.RESOLVED_ENTITY]),
        A.CHUNK_FILTERER: NlpHcFeatureNode(A.CHUNK_FILTERER, [F.DOCUMENT, F.CHUNK], [F.CHUNK]),  # TODO chunk subtype?,
        A.CHUNK_KEY_PHRASE_EXTRACTION: NlpHcFeatureNode(A.CHUNK_KEY_PHRASE_EXTRACTION, [F.DOCUMENT, F.CHUNK],
                                                        [F.CHUNK]),
        A.CHUNK_MERGE: NlpHcFeatureNode(A.CHUNK_MERGE, [F.CHUNK, F.CHUNK], [F.CHUNK]),
        A.CONTEXTUAL_PARSER: NlpHcFeatureNode(A.CONTEXTUAL_PARSER, [F.DOCUMENT, F.TOKEN], [F.CHUNK]),
        A.DE_IDENTIFICATION: NlpHcFeatureNode(A.DE_IDENTIFICATION, [F.DOCUMENT, F.TOKEN, F.NAMED_ENTITY_CONVERTED],
                                              [F.DOCUMENT_DE_IDENTIFIED]),
        A.TRAINABLE_DE_IDENTIFICATION: NlpHcFeatureNode(A.DE_IDENTIFICATION, [F.DOCUMENT, F.TOKEN, F.CHUNK],
                                                        [F.DOCUMENT]),
        A.DOCUMENT_LOG_REG_CLASSIFIER: NlpHcFeatureNode(A.DOCUMENT_LOG_REG_CLASSIFIER, [F.TOKEN],
                                                        [F.DOCUMENT_CLASSIFICATION]),
        A.TRAINABLE_DOCUMENT_LOG_REG_CLASSIFIER: NlpHcFeatureNode(A.TRAINABLE_DOCUMENT_LOG_REG_CLASSIFIER, [F.TOKEN],
                                                                  [F.DOCUMENT_CLASSIFICATION]),
        A.DRUG_NORMALIZER: NlpHcFeatureNode(A.DRUG_NORMALIZER, [F.DOCUMENT], [F.DOCUMENT_NORMALIZED]),
        # A.# FEATURES_ASSEMBLER : NlpHcFeatureNode( [H_F.FEATURE_VECTOR]) # TODO data types?,
        A.GENERIC_CLASSIFIER: NlpHcFeatureNode(A.GENERIC_CLASSIFIER, [H_F.FEATURE_VECTOR], [F.DOCUMENT_CLASSIFICATION]),
        A.TRAINABLE_GENERIC_CLASSIFIER: NlpHcFeatureNode(A.TRAINABLE_GENERIC_CLASSIFIER, [H_F.FEATURE_VECTOR],
                                                         [F.DOCUMENT_CLASSIFICATION]),
        A.IOB_TAGGER: NlpHcFeatureNode(A.IOB_TAGGER, [F.TOKEN, F.CHUNK], [F.NAMED_ENTITY_IOB]),
        A.MEDICAL_NER: NlpHcFeatureNode(A.MEDICAL_NER, [F.DOCUMENT, F.TOKEN, F.WORD_EMBEDDINGS], [F.NAMED_ENTITY_IOB]),
        A.TRAINABLE_MEDICAL_NER: NlpHcFeatureNode(A.TRAINABLE_MEDICAL_NER, [F.DOCUMENT, F.TOKEN, F.WORD_EMBEDDINGS],
                                                  [F.NAMED_ENTITY_IOB]),
        A.NER_CHUNKER: NlpHcFeatureNode(A.NER_CHUNKER, [F.DOCUMENT, F.NAMED_ENTITY_IOB], [F.CHUNK]),
        A.NER_CONVERTER_INTERNAL: NlpHcFeatureNode(A.NER_CONVERTER_INTERNAL, [F.DOCUMENT, F.TOKEN, F.NAMED_ENTITY_IOB],
                                                   [F.NAMED_ENTITY_CONVERTED]),
        A.NER_DISAMBIGUATOR: NlpHcFeatureNode(A.NER_DISAMBIGUATOR, [F.CHUNK, F.SENTENCE_EMBEDDINGS],
                                              [H_F.DISAMBIGUATION]),
        A.RELATION_NER_CHUNKS_FILTERER: NlpHcFeatureNode(A.RELATION_NER_CHUNKS_FILTERER,
                                                         [F.CHUNK, F.UNLABLED_DEPENDENCY],
                                                         [F.CHUNK]),
        A.RE_IDENTIFICATION: NlpHcFeatureNode(A.RE_IDENTIFICATION, [F.DOCUMENT, F.CHUNK], [F.DOCUMENT_RE_IDENTIFIED]),
        A.RELATION_EXTRACTION: NlpHcFeatureNode(A.RELATION_EXTRACTION,
                                                [F.NAMED_ENTITY_CONVERTED, F.WORD_EMBEDDINGS, F.POS,
                                                 F.UNLABLED_DEPENDENCY],
                                                [H_F.RELATION]),

        A.ZERO_SHOT_RELATION_EXTRACTION: NlpHcFeatureNode(A.ZERO_SHOT_RELATION_EXTRACTION,
                                                          [F.NAMED_ENTITY_CONVERTED, F.DOCUMENT, ],
                                                          [H_F.RELATION]),

        A.TRAINABLE_RELATION_EXTRACTION: NlpHcFeatureNode(A.TRAINABLE_RELATION_EXTRACTION,
                                                          [F.NAMED_ENTITY_CONVERTED, F.WORD_EMBEDDINGS, F.POS,
                                                           F.UNLABLED_DEPENDENCY],
                                                          [H_F.RELATION]),
        A.RELATION_EXTRACTION_DL: NlpHcFeatureNode(A.RELATION_EXTRACTION_DL,
                                                   [F.NAMED_ENTITY_CONVERTED, F.DOCUMENT], [H_F.RELATION]),
        A.SENTENCE_ENTITY_RESOLVER: NlpHcFeatureNode(A.SENTENCE_ENTITY_RESOLVER,
                                                     [F.DOCUMENT_FROM_CHUNK, F.SENTENCE_EMBEDDINGS],
                                                     [H_F.RESOLVED_ENTITY]),
        A.TRAINABLE_SENTENCE_ENTITY_RESOLVER: NlpHcFeatureNode(A.TRAINABLE_SENTENCE_ENTITY_RESOLVER,
                                                               [F.SENTENCE_EMBEDDINGS], [H_F.ASSERTION]),
        A.MEDICAL_BERT_FOR_TOKEN_CLASSIFICATION: NlpFeatureNode(A.MEDICAL_BERT_FOR_TOKEN_CLASSIFICATION,
                                                                [F.DOCUMENT, F.TOKEN],
                                                                [F.TOKEN_CLASSIFICATION]),

        A.MEDICAL_BERT_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(A.MEDICAL_BERT_FOR_SEQUENCE_CLASSIFICATION,
                                                                   [F.DOCUMENT, F.TOKEN],
                                                                   [F.SEQUENCE_CLASSIFICATION]),
        A.MEDICAL_DISTILBERT_FOR_SEQUENCE_CLASSIFICATION: NlpFeatureNode(
            A.MEDICAL_DISTILBERT_FOR_SEQUENCE_CLASSIFICATION, [F.DOCUMENT, F.TOKEN],
            [F.SEQUENCE_CLASSIFICATION]),

    }



class TokenBertHealthcare:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import MedicalBertForTokenClassifier
        return MedicalBertForTokenClassifier.pretrained() \
            .setInputCols("sentence", "token") \
            .setOutputCol("ner")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import MedicalBertForTokenClassifier
        return MedicalBertForTokenClassifier.pretrained(name, language, bucket) \
            .setInputCols("sentence", "token") \
            .setOutputCol("ner")



from sparknlp.annotator import *

class SparkNLPDeepSentenceDetector:
    @staticmethod
    def get_default_model():
        return SentenceDetectorDLModel.pretrained() \
            .setInputCols(["document"]) \
            .setOutputCol("sentence")






from nlu.pipe.extractors.extractor_methods.base_extractor_methods import *
from nlu.pipe.extractors.extractor_methods.helper_extractor_methods import *

"""
This file contains methods to get pre-defined configurations for every annotator.
Extractor_resolver.py should be used to resolve SparkNLP Annotator classes to methods 
in this file, which return the corrosponding configs that need to be passed to 
the master_extractor() call.

This file is where all the in extractor_base_data_classes.py Dataclasses are combined with the 
extractors defined in extractor_methods.py.

"""


def default_get_nothing(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        name='nothing_extractor',
        description='Extracts nothing. Useful for annotators with irrelevant data'
    )


def default_only_result_config(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        name='Default result extractor',
        description='Just gets the result field'
    )


def default_full_config(output_col_prefix='DEFAULT'):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_positions=True,
        get_begin=True,
        get_end=True,
        get_embeds=True,
        get_result=True,
        get_meta=True,
        get_full_meta=True,
        get_annotator_type=True,
        name='default_full',
        description='Default full configuration, keeps all data and gets all metadata fields',

    )


def default_NER_converter_licensed_config(output_col_prefix='entities'):
    """Extracts NER tokens withouth positions, just the converted IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        get_meta=True,
        meta_white_list=['entity', 'confidence','sentence', 'chunk'],
        name='default_ner',
        description='Converts IOB-NER representation into entity representation and generates confidences for the entire entity chunk',
    )



def default_chunk_mapper_config(output_col_prefix='mapped_entity'):
    """Extracts NER tokens withouth positions, just the converted IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        get_meta=True,
        meta_white_list=['relation', 'all_relations','chunk', 'entity',
                         'sentence'
                         ],  # MAYBE DROP 'chunk', 'entity'default_chunk_mapper_config, sentence
        name='default_ner',
        meta_data_extractor=SparkNLPExtractor(extract_chunk_mapper_relation_data,
                                          'Get ChunkMapper Relation Metadata',
                                          'Get ChunkMapper Relation Metadata'),
        description='Extract Chunk Mapper with relation Data',
    )



def default_chunk_resolution_config(output_col_prefix='resolved_entities'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        get_meta=True,
        meta_white_list=['confidence', 'resolved_text'],  # sentence, chunk
        name='default_ner',

    )


def full_resolver_config(output_col_prefix='DEFAULT'):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_positions=True,
        get_begin=True,
        get_end=True,
        get_embeds=True,
        get_result=True,
        get_meta=True,
        get_full_meta=True,
        get_annotator_type=True,
        name='default_full',
        description='Full resolver outputs, with any _k_ field in the metadata dict splitted :::',
        meta_data_extractor=SparkNLPExtractor(extract_resolver_all_k_subfields_splitted,
                                              'Splits all _k_ fields on :::d returns all other fields as corrosponding to pop config',
                                              'split all _k_ fields')

    )


def resolver_conifg_with_metadata(output_col_prefix='DEFAULT'):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_meta=True,
        get_result=True,
        get_full_meta=True,
        name='with metadata',
        description='Full resolver outputs, with any _k_ field in the metadata dict splitted :::',
        meta_data_extractor=SparkNLPExtractor(extract_resolver_all_k_subfields_splitted,
                                              'Splits all _k_ fields on :::d returns all other fields as corrosponding to pop config',
                                              'split all _k_ fields')

    )


def default_relation_extraction_positional_config(output_col_prefix='extracted_relations'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        get_meta=True,
        get_full_meta=True,
        name='full_relation_extraction',
        description='Get relation extraction result and all metadata, with positions of entities',
    )


def default_relation_extraction_config(output_col_prefix='extracted_relations'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        meta_white_list=[],
        get_meta=True,
        meta_black_list=['entity1_begin', 'entity2_begin', 'entity1_end', 'entity2_end', ],
        name='default_relation_extraction',
        description='Get relation extraction result and all metadata, positions of entities excluded',
    )


def default_de_identification_config(output_col_prefix='de_identified'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        name='positional_relation_extraction',
        description='Get relation extraction result and all metadata, which will include positions of entities chunks',
    )


def default_assertion_config(output_col_prefix='assertion'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        name='default_assertion_extraction',
        get_meta=True,
        meta_white_list=['confidence'],
        description='Gets the assertion result and confidence',
    )


def default_ner_config(output_col_prefix='med_ner'):
    return default_only_result_config(output_col_prefix)


def default_ner_config(output_col_prefix='med_ner'):
    return default_get_nothing(output_col_prefix)


def default_feature_assembler_config(output_col_prefix='feature_assembler'):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=False,
        name='features_assembled',
        get_meta=False,
        description='Gets nothing',
    )


def default_generic_classifier_config(output_col_prefix='generic_classifier'):
    return SparkNLPExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_result=True,
        name='generic_classifier',
        get_meta=True,
        meta_white_list=['confidence'],
        description='Gets the  result and confidence',
    )



class Binary2Image:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import BinaryToImage
        return BinaryToImage() \
            .setInputCol("content") \
            .setOutputCol("image")



from sparknlp.annotator import *
from nlu.pipe.viz.vis_utils_OS import VizUtilsOS
import random


class VizUtils():
    """Utils for interfacing with the Spark-NLP-Display lib"""

    @staticmethod
    def get_random():
        return random.randint(0, 1333333333337)

    @staticmethod
    def infer_viz_type(pipe) -> str:
        """For a given NLUPipeline, infers which visualizations are applicable. """
        if pipe.has_licensed_components:
            from nlu.pipe.viz.vis_utils_HC import VizUtilsHC
            return VizUtilsHC.infer_viz_licensed(pipe)
        else:
            return VizUtilsOS.infer_viz_open_source(pipe)

    @staticmethod
    def viz_OS(anno_res, pipe, viz_type, viz_colors, labels_to_viz, is_databricks_env, write_to_streamlit,
               streamlit_key,
               ner_col, pos_col, dep_untyped_col, dep_typed_col):
        """Vizualize open source component_to_resolve"""
        streamlit_key = VizUtils.get_random() if streamlit_key == "RANDOM" else streamlit_key
        if viz_type == 'ner':
            return VizUtilsOS.viz_ner(anno_res, pipe, labels_to_viz, viz_colors, is_databricks_env, write_to_streamlit,
                                      streamlit_key, ner_col)
        elif viz_type == 'dep':
            return VizUtilsOS.viz_dep(anno_res, pipe, is_databricks_env, write_to_streamlit, streamlit_key,
                                      pos_col, dep_untyped_col, dep_typed_col)
        else:
            raise ValueError(
                "Could not find applicable viz_type. Please make sure you specify either ner, dep, resolution, relation, assert or dep and have loaded corrosponding components")

    @staticmethod
    def viz_HC(anno_res, pipe, viz_type, viz_colors, labels_to_viz, is_databricks_env, write_to_streamlit,
               ner_col, pos_col, dep_untyped_col, dep_typed_col, resolution_col, relation_col, assertion_col):
        """Vizualize licensed component_to_resolve"""
        from nlu.pipe.viz.vis_utils_HC import VizUtilsHC
        if viz_type == 'ner':
            return VizUtilsHC.viz_ner(anno_res, pipe, labels_to_viz, viz_colors, is_databricks_env, write_to_streamlit,
                                      ner_col)
        elif viz_type == 'dep':
            return VizUtilsHC.viz_dep(anno_res, pipe, is_databricks_env, write_to_streamlit, dep_untyped_col,
                                      dep_typed_col, pos_col)
        elif viz_type == 'resolution':
            return VizUtilsHC.viz_resolution(anno_res, pipe, viz_colors, is_databricks_env,
                                             write_to_streamlit, ner_col, resolution_col)
        elif viz_type == 'relation':
            return VizUtilsHC.viz_relation(anno_res, pipe, is_databricks_env, write_to_streamlit, relation_col, )
        elif viz_type == 'assert':
            return VizUtilsHC.viz_assertion(anno_res, pipe, viz_colors, is_databricks_env, write_to_streamlit, ner_col,
                                            assertion_col)
        else:
            raise ValueError(
                "Could not find applicable viz_type. Please make sure you specify either ner, dep, resolution, relation, assert or dep and have loaded corrosponding components")


"""Define whiche annotators model_anno_obj are definable by which vizualizer. There are 5 in total, 2 open source and 5 HC"""
# vizalbe_components_OC = {
#     'ner' : [NerConverter],
#     'dep' : [DependencyParserModel],
# }

# vizalbe_components_HC = {
#     'ner':[NerConverter,NerConverterInternal],
#     'resolution' : [SentenceEntityResolverModel, ChunkEntityResolverModel] ,
#     'relation'   : [RelationExtractionModel,RelationExtractionDLModel],
#     'assert'     : [AssertionDLModel,AssertionLogRegApproach],
#     'dep' : [DependencyParserModel],
# }



from sparknlp.annotator import MPNetEmbeddings


class MPNetSentence:
    @staticmethod
    def get_default_model():
        return MPNetEmbeddings.pretrained() \
            .setInputCols(["documents"]) \
            .setOutputCol("mpnet_embeddings")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return MPNetEmbeddings.pretrained(name,language,bucket) \
            .setInputCols(["documents"]) \
            .setOutputCol("mpnet_embeddings")






from sparknlp.annotator import *


class TapasQA:
    @staticmethod
    def get_default_model():
        return TapasForQuestionAnswering.pretrained() \
            .setInputCols('table', "document") \
            .setOutputCol("answer")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return TapasForQuestionAnswering.pretrained(name, language, bucket) \
            .setInputCols('table', "document") \
            .setOutputCol("answer")



class ImageSplitRegions:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import ImageSplitRegions
        return ImageSplitRegions() \
            .setInputCol("ocr_image") \
            .setInputRegionsCol("region") \
            .setOutputCol("image_region")

# .setInputRegionsCol("ocr_table_16969+
#
#
#
# ") \


from sparknlp.annotator import *

class ChunkEmbedder:
    @staticmethod
    def get_default_model():
        return ChunkEmbeddings() \
            .setInputCols(["entities", "word_embeddings"]) \
            .setOutputCol("chunk_embeddings") \
            .setPoolingStrategy("AVERAGE")


class ChunkMapper:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import ChunkMapperModel
        return ChunkMapperModel() \
            .setInputCols("chunk") \
            .setOutputCol("mapped_chunk")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import ChunkMapperModel
        return ChunkMapperModel.pretrained(name, language, bucket) \
            .setInputCols('mapped_chunk') \
            .setOutputCol("mapped_chunk")



"""
The FeaturesAssembler is used to collect features from different columns. 
It can collect features from single value columns (anything which can be cast to a float, if casts fails then the value is set to 0),
 array columns or SparkNLP annotations (if the annotation is an embedding, it takes the embedding, otherwise tries to cast the result field). 
The output of the transformer is a FEATURE_VECTOR annotation (the numeric vector is in the embeddings field).
"""
class SparkNLPFeatureAssembler:
    @staticmethod

    def get_default_model():
        from sparknlp_jsl.base import FeaturesAssembler
        return FeaturesAssembler() \
            .setInputCols(["%%%feature_elements%%%"]) \
            .setOutputCol("feature_vector")



from sparknlp.annotator import *

class SentenceDetectorDeep:
    @staticmethod
    def get_default_model():
        return SentenceDetectorDLModel\
            .pretrained()\
            .setInputCols(["document"]) \
            .setOutputCol("sentence")


    @staticmethod
    def get_pretrained_model(name,lang, bucket=None):
        return SentenceDetectorDLModel.pretrained(name,lang,bucket) \
            .pretrained() \
            .setInputCols(["document"]) \
            .setOutputCol("sentence")
    #
    #
    # @staticmethod
    # def get_trainable_model():
    #     return SentenceDetectorDLApproach \
    #         .setInputCol("document") \
    #         .setOutputCol("sentence")



import json
import logging
from typing import Union

import sparknlp
from pyspark.sql.types import StructType, StructField, StringType
from sparknlp.base import *
from sparknlp.base import LightPipeline

from nlu.pipe.col_substitution.col_name_substitution_utils import ColSubstitutionUtils
from nlu.pipe.extractors.extractor_configs_HC import default_full_config
from nlu.pipe.extractors.extractor_methods.base_extractor_methods import *
from nlu.pipe.extractors.extractor_methods.ocr_extractors import extract_tables
from nlu.pipe.nlu_component import NluComponent
from nlu.pipe.utils.component_utils import ComponentUtils
from nlu.pipe.utils.data_conversion_utils import DataConversionUtils
from nlu.pipe.utils.output_level_resolution_utils import OutputLevelUtils
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from nlu.universe.feature_node_ids import NLP_NODE_IDS
from nlu.universe.universes import Licenses
from nlu.utils.environment.env_utils import is_running_in_databricks, try_import_streamlit

logger = logging.getLogger('nlu')


class NLUPipeline(dict):
    # we inherhit from dict so the component_list is indexable and we have a nice shortcut for accessing the spark nlp model_anno_obj
    def __init__(self):
        """ Initializes a pretrained pipeline, should only be created after a
        Spark Context has been created
          """
        self.spark = sparknlp.start()
        self.provider = 'sparknlp'
        self.pipe_ready = False  # ready when we have created a spark df
        self.failed_pyarrow_conversion = False
        self.anno2final_cols = []  # Maps Anno to output pandas col
        self.contains_ocr_components = False
        self.contains_audio_components = False
        self.has_nlp_components = False
        self.nlu_ref = ''
        self.raw_text_column = 'text'
        self.raw_text_matrix_slice = 1  # place holder for getting text from matrix
        self.spark_nlp_pipe = None
        self.has_trainable_components = False
        self.needs_fitting = True
        self.is_fitted = False
        self.output_positions = False  # Wether to putput positions of Features in the final output. E.x. positions of tokens, entities, dependencies etc.. inside of the input document.
        self.prediction_output_level = ''  # either document, chunk, sentence, token
        self.component_output_level = ''  # document or sentence, depending on how input dependent Sentence/Doc classifier are fed
        self.output_different_levels = True
        self.light_pipe_configured = False
        self.components = []  # orderd list of nlu_component objects
        self.output_datatype = 'pandas'  # What data type should be returned after predict either spark, pandas, modin, numpy, string or array
        self.lang = 'en'
        self.vanilla_transformer_pipe = None
        self.estimator_pipe = None
        self.light_transformer_pipe = None
        self.has_licensed_components = False
        self.has_span_classifiers = False
        self.prefer_light = False
        self.has_table_qa_models = False
        self.requires_image_format = False
        self.requires_binary_format = False
        self.is_light_pipe_incompatible = False
    def add(self, component: NluComponent, nlu_reference=None, pretrained_pipe_component=False,
            name_to_add='', idx=None):
        '''

        :param component:
        :return: None
        '''
        if idx:
            self.components.insert(idx, component)
        else:
            self.components.append(component)
        # ensure that input/output cols are properly set
        # Spark NLP model_anno_obj reference shortcut
        name = component.name  # .replace(' ', '').replace('train.', '')

        if StorageRefUtils.has_storage_ref(component) and component.is_trained:
            # Converters have empty storage ref intially
            storage_ref = StorageRefUtils.extract_storage_ref(component)
            if storage_ref != '':
                name = name + '@' + StorageRefUtils.extract_storage_ref(component)
        logger.info(f"Adding {name} to internal component_list")

        # Configure output column names of classifiers from category to something more meaningful
        # if self.isInstanceOfNlpClassifer(component_to_resolve.model_anno_obj): self.configure_outputs(component_to_resolve, nlu_ref)

        if name_to_add == '':
            # Add Component as self.index and in attributes
            if component.is_storage_ref_producer and component.nlu_ref not in self.keys() and not pretrained_pipe_component:
                self[name] = component.model
            elif name not in self.keys():
                self[name] = component.model
            else:
                nlu_identifier = ComponentUtils.get_nlu_ref_identifier(component)
                self[name + "@" + nlu_identifier] = component.model
        else:
            self[name_to_add] = component.model

    def get_sample_spark_dataframe(self):
        data = {"text": ['This day sucks', 'I love this day', 'I dont like Sami']}
        text_df = pd.DataFrame(data)
        return sparknlp.start().createDataFrame(data=text_df)

    def verify_all_labels_exist(self, dataset):
        return 'y' in dataset.columns  # or 'label' in dataset.columns or 'labels' in dataset.columns

    def fit(self, dataset=None, dataset_path=None, label_seperator=','):
        '''
        if dataset is  string with '/' in it, its dataset path!
        Converts the input Pandas Dataframe into a Spark Dataframe and trains a model_anno_obj on it.
        :param dataset: Pandas dataset to train on, should have a y column for label and 'text' column for text features
        :param dataset_path: Path to a CONLL2013 format dataset. It will be read for NER and POS training.
        :param label_seperator: If multi_classifier is trained, this seperator is used to split the elements into an Array column for Pyspark
        :return: A nlu pipeline with models fitted.
        '''
        from nlu.pipe.pipe_logic import PipeUtils

        stages = []
        for component in self.components:
            stages.append(component.model)
        self.spark_estimator_pipe = Pipeline(stages=stages)

        if dataset_path != None and 'ner' in self.nlu_ref:
            from sparknlp.training import CoNLL
            s_df = CoNLL().readDataset(self.spark, path=dataset_path, )
            self.vanilla_transformer_pipe = self.spark_estimator_pipe.fit(s_df.withColumnRenamed('label', 'y'))
            self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe)
        elif dataset_path != None and 'pos' in self.nlu_ref:
            from sparknlp.training import POS
            s_df = POS().readDataset(self.spark, path=dataset_path, delimiter=label_seperator, outputPosCol="y",
                                     outputDocumentCol="document", outputTextCol="text")
            self.vanilla_transformer_pipe = self.spark_estimator_pipe.fit(s_df)
            self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe)

        elif isinstance(dataset, pd.DataFrame) and 'multi' in self.nlu_ref:
            schema = StructType([
                StructField("y", StringType(), True),
                StructField("text", StringType(), True)
            ])
            from pyspark.sql import functions as F
            df = self.spark.createDataFrame(data=dataset).withColumn('y', F.split('y', label_seperator))
            # df = self.spark.createDataFrame(data=dataset, schema=schema).withColumn('y',F.split('y',label_seperator))
            # df = self.spark.createDataFrame(dataset)
            self.vanilla_transformer_pipe = self.spark_estimator_pipe.fit(df)
            self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe)

        elif isinstance(dataset, pd.DataFrame):
            if not self.verify_all_labels_exist(dataset): raise ValueError(
                f"Could not detect label in provided columns={dataset.columns}\nMake sure a column named label, labels or y exists in your dataset.")
            dataset.y = dataset.y.apply(str)
            if self.has_licensed_components:
                # Configure Feature Assembler
                for c in self.components:
                    if c.name == 'feature_assembler':
                        vector_assembler_input_cols = [c for c in dataset.columns if
                                                       c != 'text' and c != 'y' and c != 'label' and c != 'labels']
                        c.model.setInputCols(vector_assembler_input_cols)
                        # os_components.model_anno_obj.spark_input_column_names = vector_assembler_input_cols
                # Configure Chunk resolver Sentence to Document substitution in all cols. when training Chunk resolver, we must substitute all SENTENCE cols with DOC. We MAY NOT FEED SENTENCE to CHUNK RESOLVE or we get errors
                self.components = PipeUtils.configure_component_output_levels_to_document(self)

            # Substitute @ notation to ___ because it breaks Pyspark SQL Parser...
            for c in self.components:
                for inp in c.spark_input_column_names:
                    if 'chunk_embedding' in inp:
                        c.spark_input_column_names.remove(inp)
                        c.spark_input_column_names.append(inp.replace('@', "___"))
                        c.model.setInputCols(c.spark_input_column_names)
                    if 'sentence_embedding' in inp:
                        c.spark_input_column_names.remove(inp)
                        c.spark_input_column_names.append(inp.replace('@', "___"))
                        c.model.setInputCols(c.spark_input_column_names)
                for out in c.spark_output_column_names:
                    if 'chunk_embedding' in out:
                        c.spark_output_column_names.remove(out)
                        c.spark_output_column_names.append(out.replace('@', "___"))
                        c.model.setOutputCol(c.spark_output_column_names[0])
                    if 'sentence_embedding' in out:
                        c.spark_output_column_names.remove(out)
                        c.spark_output_column_names.append(out.replace('@', "___"))
                        c.model.setOutputCol(c.spark_output_column_names[0])

            stages = []
            for component in self.components: stages.append(component.model)
            ## TODO set storage ref on fitted model_anno_obj
            self.spark_estimator_pipe = Pipeline(stages=stages)
            self.vanilla_transformer_pipe = self.spark_estimator_pipe.fit(
                DataConversionUtils.pdf_to_sdf(dataset, self.spark)[0])
            self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe)

        elif isinstance(dataset, pyspark.sql.DataFrame):
            if not self.verify_all_labels_exist(dataset): raise ValueError(
                f"Could not detect label in provided columns={dataset.columns}\nMake sure a column named label, labels or y exists in your dataset.")
            self.vanilla_transformer_pipe = self.spark_estimator_pipe.fit(
                DataConversionUtils.sdf_to_sdf(dataset, self.spark))
            self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe)

        else:
            # fit on empty dataframe since no data provided
            if not self.is_fitted:
                logger.info(
                    'Fitting on empty Dataframe, could not infer correct training method. This is intended for non-trainable pipelines.')
                self.vanilla_transformer_pipe = self.spark_estimator_pipe.fit(self.get_sample_spark_dataframe())
                if not self.is_light_pipe_incompatible:
                    self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe)

        self.has_trainable_components = False
        self.is_fitted = True
        self.light_pipe_configured = True
        self.components = PipeUtils.replace_untrained_component_with_trained(self, self.vanilla_transformer_pipe)

        return self

    def get_extraction_configs(self, full_meta, positions, get_embeddings, processed):
        """Search first OC namespace and if not found the HC Namespace for each Annotator Class in pipeline and get
        corresponding config Returns a dictionary of methods, where keys are column names values are methods  that
        are applied to extract and represent the data in these columns in a more pythonic and panda-esque way
        """
        c_level_mapping = OutputLevelUtils.get_output_level_mapping_by_component(self)
        # todo doc level annos and same level annos can be popped always.
        anno_2_ex_config = {}
        for c in self.components:
            if c.license == Licenses.ocr:
                # OCR can output more than 1 col
                extractors = {col: c.pdf_extractor_methods['default'](output_col_prefix=col) for col in
                              c.spark_output_column_names}
                anno_2_ex_config.update(extractors)
                continue
            if 'embedding' in c.type and not get_embeddings:
                continue
            if c.name == NLP_NODE_IDS.FINISHER:
                anno_2_ex_config = {**anno_2_ex_config, **self.__get_finisher_conf(c)}
                continue
            for col in c.spark_output_column_names:
                if 'default' in c.pdf_extractor_methods.keys() and not full_meta:
                    anno_2_ex_config[col] = c.pdf_extractor_methods['default'](output_col_prefix=col)
                elif 'default_full' in c.pdf_extractor_methods.keys() and full_meta:
                    anno_2_ex_config[col] = c.pdf_extractor_methods['default_full'](output_col_prefix=col)
                else:
                    # Fallback if no output defined
                    anno_2_ex_config[col] = default_full_config(output_col_prefix=col)

                # Tune the Extractor configs based on prediction parameters
                if c_level_mapping[c] == 'document' and not anno_2_ex_config[col].pop_never:
                    # Disable popping for doc level outputs, output will not be [element] but instead element in each row.
                    anno_2_ex_config[col].pop_meta_list = True
                    anno_2_ex_config[col].pop_result_list = True
                if positions:
                    anno_2_ex_config[col].get_positions = True
                else:
                    anno_2_ex_config[col].get_begin = False
                    anno_2_ex_config[col].get_end = False
                    anno_2_ex_config[col].get_positions = False
                if c.loaded_from_pretrained_pipe:
                    # Use original col name of pretrained pipes as prefix
                    anno_2_ex_config[col].output_col_prefix = col
        # drop all entries which are missing in DF
        # Finisher Annotator could have dropped them
        bad_c = []
        for c in anno_2_ex_config.keys():
            if c not in processed.columns:
                bad_c.append(c)
        for c in bad_c:
            del anno_2_ex_config[c]

        return anno_2_ex_config

    def unpack_and_apply_extractors(self, pdf: Union[pyspark.sql.DataFrame, pd.DataFrame], keep_stranger_features=True,
                                    stranger_features=[],
                                    anno_2_ex_config={},
                                    light_pipe_enabled=True,
                                    get_embeddings=False
                                    ) -> pd.DataFrame:
        """1. Unpack SDF to PDF with Spark NLP Annotator Dictionaries
           2. Get the extractor configs for the corresponding Annotator classes
           3. Apply The extractor configs with the extractor methods to each column and merge back with zip/explode
           Uses optimized PyArrow conversion to avoid representing data multiple times between the JVM and PVM

           Can process Spark DF output from Vanilla pipes and Pandas Converts outputs of Lightpipeline
           """
        # Light pipe, does not fetch emebddings
        if light_pipe_enabled and not get_embeddings and not isinstance(pdf,
                                                                        pyspark.sql.dataframe.DataFrame) or self.prefer_light:
            return apply_extractors_and_merge(extract_light_pipe_rows(pdf),
                                              anno_2_ex_config, keep_stranger_features, stranger_features)

        # Vanilla Spark Pipe
        return apply_extractors_and_merge(pdf.toPandas().applymap(extract_pyspark_rows), anno_2_ex_config,
                                          keep_stranger_features, stranger_features)

    def pythonify_spark_dataframe(self, processed,
                                  keep_stranger_features=True,
                                  stranger_features=[],
                                  drop_irrelevant_cols=True,
                                  output_metadata=False,
                                  positions=False,
                                  output_level='',
                                  get_embeddings=True):
        '''
        This functions takes in a spark dataframe with Spark NLP annotations in it and transforms it into a Pandas
        Dataframe with common feature types for further NLP/NLU downstream tasks. It will recycle Indexes from Pandas
        DataFrames and Series if they exist, otherwise a custom id column will be created which is used as index later
        on It does this by performing the following consecutive steps : 1. Select columns to explode 2. Select
        columns to keep 3. Rename columns 4. Create Pandas Dataframe object
        :param processed: Spark dataframe which an NLU pipeline has transformed
        :param output_level: The output level at which returned pandas Dataframe should be
        :param keep_stranger_features : Whether to keep additional features from the input DF when generating the output
                                        DF or if they should be discarded for the final output DF
        :param stranger_features: A list of features which are not known to NLU and inside the input DF.
                                    Basically all columns, which are not named 'text' in the input.
                                    If keep_stranger_features== True, then these features will be exploded,
                                    if pipe_prediction_output_level == document, otherwise they will not be exploded
        :param output_metadata: Whether to keep or drop additional metadata or predictions, like prediction confidence
        :return: Pandas dataframe which easy accessible features
        '''
        from nlu.pipe.pipe_logic import PipeUtils

        if PipeUtils.has_table_extractor(self):
            # If pipe has table extractors, we return list of tables or table itself if only one detected
            processed = extract_tables(processed)
            if len(processed) == 1:
                return processed[0]
            return processed

        stranger_features += ['origin_index']
        if output_level == '':
            # Infer output level if none defined
            self.prediction_output_level = OutputLevelUtils.infer_prediction_output_level(self)
            logger.info(f'Inferred and set output level of pipeline to {self.prediction_output_level}')
        else:
            self.prediction_output_level = output_level

        # Get mapping from component to feature extractor method configs
        col_2_ex_config = self.get_extraction_configs(output_metadata, positions, get_embeddings, processed)

        # Processed becomes pandas after applying extractors
        processed = self.unpack_and_apply_extractors(processed, keep_stranger_features, stranger_features,
                                                     col_2_ex_config, self.light_pipe_configured, get_embeddings)

        # Get mapping between column_name and pipe_prediction_output_level
        same_level = OutputLevelUtils.get_columns_at_same_level_of_pipe(self, processed, col_2_ex_config,
                                                                        get_embeddings)
        logger.info(f"Extracting for same_level_cols = {same_level}\n")
        processed = zip_and_explode(processed, same_level)
        processed = self.convert_embeddings_to_np(processed)
        processed = ColSubstitutionUtils.substitute_col_names(processed, col_2_ex_config, self, stranger_features,
                                                              get_embeddings)
        processed = processed.loc[:, ~processed.columns.duplicated()]

        if drop_irrelevant_cols and not output_metadata:
            processed = processed[self.drop_irrelevant_cols(list(processed.columns))]
        # Sort cols alphabetically
        processed = processed.reindex(sorted(processed.columns), axis=1)
        return processed

    def convert_embeddings_to_np(self, pdf):
        '''
        convert all the columns in a pandas df to numpy
        :param pdf: Pandas Dataframe whose embedding column will be converted to numpy array objects
        :return:
        '''

        for col in pdf.columns:
            if 'embed' in col:
                pdf[col] = pdf[col].apply(lambda x: np.array(x))
        return pdf

    def finalize_return_datatype(self, df):
        '''
        Take in a Spark dataframe with only relevant columns remaining.
        Depending on what value is set in self.output_datatype, this method will cast the final SDF into Pandas/Spark/Numpy/Modin/List objects
        :param df:
        :return: The predicted Data as datatype dependign on self.output_datatype
        '''
        if self.output_datatype == 'spark':
            return df
        elif self.output_datatype == 'pandas':
            return df
        elif self.output_datatype == 'modin':
            import modin.pandas as mpd
            return mpd.DataFrame(df)
        elif self.output_datatype == 'pandas_series':
            return df
        elif self.output_datatype == 'modin_series':
            import modin.pandas as mpd
            return mpd.DataFrame(df)
        elif self.output_datatype == 'numpy':
            return df.to_numpy()
        return df

    def drop_irrelevant_cols(self, cols, keep_origin_index=False):
        '''
        Takes in a list of column names removes the elements which are irrelevant to the current output level.
        This will be run before returning the final df
        Drop column candidates are document, sentence, token, chunk.
        columns which are NOT AT THE SAME output level will be dropped
        :param cols:  list of column names in the df
        :return: list of columns with the irrelevant names removed
        '''
        if 'doc2chunk' in cols: cols.remove('doc2chunk')
        if 'doc2chunk' in cols: cols.remove('doc2chunk')

        if self.prediction_output_level == 'token':
            if 'document' in cols: cols.remove('document')
            if 'chunk' in cols: cols.remove('chunk')
            if 'sentence' in cols: cols.remove('sentence')
        if self.prediction_output_level == 'sentence':
            if 'token' in cols: cols.remove('token')
            if 'chunk' in cols: cols.remove('chunk')
            if 'document' in cols: cols.remove('document')
        if self.prediction_output_level == 'chunk':
            # if 'document' in cols: cols.remove('document')
            if 'token' in cols: cols.remove('token')
            if 'sentence' in cols: cols.remove('sentence')
        if self.prediction_output_level == 'document':
            if 'token' in cols: cols.remove('token')
            if 'chunk' in cols: cols.remove('chunk')
            if 'sentence' in cols: cols.remove('sentence')
        if self.prediction_output_level == 'relation':
            if 'token' in cols: cols.remove('token')
            if 'chunk' in cols: cols.remove('chunk')
            if 'sentence' in cols: cols.remove('sentence')
        if keep_origin_index == False and 'origin_index' in cols: cols.remove('origin_index')
        return cols

    def save(self, path, component='entire_pipeline', overwrite=True):
        # serialize data
        data = {}
        data[0] = {'nlu_ref': self.nlu_ref}
        data['is_nlu_pipe'] = True
        for i, c in enumerate(self.components):
            data[i + 1] = {'nlu_ref': c.nlu_ref, 'nlp_ref': c.nlp_ref,
                           'loaded_from_pretrained_pipe': c.loaded_from_pretrained_pipe}

        data = json.dumps(data)
        if not self.is_fitted or not hasattr(self, 'vanilla_transformer_pipe'):
            self.fit()
            self.is_fitted = True
        # self.vanilla_transformer_pipe.extractParamMap()
        if hasattr(self, 'nlu_ref'):
            """ ATTRS TO SAVE FOR EACH COMPONENT / PIPELINE: 
            - nlp ref/nlu ref
            - is loaded_form_pipe
            """
            self.vanilla_transformer_pipe._resetUid(data)
        if component == 'entire_pipeline':
            if overwrite:
                self.vanilla_transformer_pipe.write().overwrite().save(path)
            else:
                self.vanilla_transformer_pipe.save(path)
        else:
            if overwrite:
                self[component].write().overwrite().save(path)
            else:
                self[component].save(path)

    def predict(self,
                data,
                output_level='',
                positions=False,
                keep_stranger_features=True,
                metadata=False,
                multithread=True,
                drop_irrelevant_cols=True,
                return_spark_df=False,
                get_embeddings=True
                ):
        '''
        Annotates a Pandas Dataframe/Pandas Series/Numpy Array/Spark DataFrame/Python List strings /Python String

        :param data: Data to predict on
        :param output_level: output level, either document/sentence/chunk/token
        :param positions: whether to output indexes that map predictions back to position in origin string
        :param keep_stranger_features: whether to keep columns in the dataframe that are not generated by pandas. I.e.
                when you s a dataframe with 10 columns and only one of them is named text, the returned dataframe will
                only contain the text column when set to false
        :param metadata: whether to keep additional metadata in final df or not like
                confidences of every possible class for predictions.
        :param multithread: Whether to use multithreading based light pipeline. In some cases, this may cause errors.
        :param drop_irrelevant_cols: Whether to drop cols of different output levels, i.e. when predicting token level
                and drop_irrelevant_cols = True then chunk, sentence and Doc will be dropped
        :param return_spark_df: Prediction results will be returned right after transforming with the Spark NLP pipeline
                                 This will run fully distributed in on the Spark Master, but not prettify the output dataframe
        :param get_embeddings: Whether to return embeddings or not
        :return:
        '''
        from nlu.pipe.utils.predict_helper import __predict__
        return __predict__(self, data, output_level, positions, keep_stranger_features, metadata, multithread,
                           drop_irrelevant_cols, return_spark_df, get_embeddings)
    def predict_embeds(self,
                       data,
                       multithread=True,
                       return_spark_df=False,
                       ):
        '''
        Annotates a Pandas Dataframe/Pandas Series/Numpy Array/Spark DataFrame/Python List strings /Python String abd returns List of Floats or Spark-Df, only with embeddings.
        :param data: Data to predict on
                and drop_irrelevant_cols = True then chunk, sentence and Doc will be dropped
        :param return_spark_df: Prediction results will be returned right after transforming with the Spark NLP pipeline
                                 This will run fully distributed in on the Spark Master, but not prettify the output dataframe
        :param return_spark_df: return Spark-DF and not collect all data into driver instead of returning list of float
        :param multithread: Use multithreaded Light-pipeline instead of spark-pipeline

        :return:
        '''
        from nlu.pipe.utils.predict_helper import __predict__
        return __predict__(self, data, output_level=None, positions=False, keep_stranger_features=False, metadata=False,
                           multithread=multithread,
                           drop_irrelevant_cols=True, return_spark_df=return_spark_df, get_embeddings=True,
                           embed_only=True)
    def print_info(self, minimal=True):
        '''
        Print out information about every component_to_resolve currently loaded in the component_list and their configurable parameters.
        If minimal is false, all Spark NLP Model parameters will be printed, including output/label/input cols and other attributes a NLU user should not touch. Useful for debugging.
        :return: None
        '''
        print('The following parameters are configurable for this NLU pipeline (You can copy paste the examples) :')
        # list of tuples, where first element is component_to_resolve name and second element is list of param tuples, all ready formatted for printing
        all_outputs = []
        iterable = None
        for i, component_key in enumerate(self.keys()):
            s = ">>> component_list['" + component_key + "'] has settable params:"
            p_map = self[component_key].extractParamMap()

            component_outputs = []
            max_len = 0
            for key in p_map.keys():
                if minimal:
                    if "outputCol" in key.name or "labelCol" in key.name or "inputCol" in key.name or "labelCol" in key.name or 'lazyAnnotator' in key.name or 'storageref' in key.name: continue

                # print("component_list['"+ component_key +"'].set"+ str( key.name[0].capitalize())+ key.name[1:]+"("+str(p_map[key])+")" + " | Info: " + str(key.doc)+ " currently Configured as : "+str(p_map[key]) )
                # print("Param Info: " + str(key.doc)+ " currently Configured as : "+str(p_map[key]) )

                if type(p_map[key]) == str:
                    s1 = "component_list['" + component_key + "'].set" + str(key.name[0].capitalize()) + key.name[
                                                                                                         1:] + "('" + str(
                        p_map[key]) + "') "
                else:
                    s1 = "component_list['" + component_key + "'].set" + str(key.name[0].capitalize()) + key.name[
                                                                                                         1:] + "(" + str(
                        p_map[key]) + ") "

                s2 = " | Info: " + str(key.doc) + " | Currently set to : " + str(p_map[key])
                if len(s1) > max_len: max_len = len(s1)
                component_outputs.append((s1, s2))

            all_outputs.append((s, component_outputs))

        # make strings aligned
        form = "{:<" + str(max_len) + "}"
        for o in all_outputs:
            print(o[0])  # component_to_resolve name
            for o_parm in o[1]:
                if len(o_parm[0]) < max_len:
                    print(form.format(o_parm[0]) + o_parm[1])
                else:
                    print(o_parm[0] + o_parm[1])

    def print_exception_err(self, err):
        '''Print information about exception during converting or transforming dataframe'''
        import sys
        logger.exception('Exception occured')
        e = sys.exc_info()
        print("No accepted Data type or usable columns found or applying the NLU models failed. ")
        print(
            "Make sure that the first column you pass to .predict() is the one that nlu should predict on OR rename the column you want to predict on to 'text'  ")
        print(
            "On try to reset restart Jupyter session and run the setup script again, you might have used too much memory")
        print('Full Stacktrace was', e)
        print('Additional info:')
        exc_type, exc_obj, exc_tb = sys.exc_info()
        import os
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        err = sys.exc_info()[1]
        print(str(err))
        print(
            'Stuck? Contact us on Slack! https://join.slack.com/t/spark-nlp/shared_invite/zt-lutct9gm-kuUazcyFKhuGY3_0AMkxqA')

    def viz(self, text_to_viz: str, viz_type='', labels_to_viz=None, viz_colors={}, return_html=False,
            write_to_streamlit=False, streamlit_key='NLU_streamlit',
            ner_col=None,
            pos_col=None,
            dep_untyped_col=None,
            dep_typed_col=None,
            resolution_col=None,
            relation_col=None,
            assertion_col=None,
            ):
        """Visualize predictions of a Pipeline, using Spark-NLP-Display
        text_to_viz : String to viz
        viz_type    : Viz type, one of [ner,dep,resolution,relation,assert]. If none defined, nlu will infer and apply all applicable viz
        labels_to_viz : Defines a subset of NER labels to viz i.e. ['PER'] , by default=[] which will display all labels. Applicable only for NER viz
        viz_colors  : Applicable for [ner, resolution, assert ] key = label, value=hex color, i.e. viz_colors={'TREATMENT':'#008080', 'problem':'#800080'}

        Any of the col parameters can be used to point to a specific model in the pipeline, if there are multiple candidates of the same type for visualization.
        I.e. multiple NER models. By default, the last model in pipe of applicable viz type will be used.
        """
        from nlu.utils.environment.env_utils import install_and_import_package
        install_and_import_package('spark-nlp-display', import_name='sparknlp_display')
        if self.vanilla_transformer_pipe is None: self.fit()
        is_databricks_env = is_running_in_databricks()
        if return_html: is_databricks_env = True
        # self.configure_light_pipe_usage(1, force=True)
        from nlu.pipe.viz.vis_utils import VizUtils

        if viz_type == '': viz_type = VizUtils.infer_viz_type(self)
        # anno_res = self.vanilla_transformer_pipe.fullAnnotate(text_to_viz)[0]
        # anno_res = self.spark.createDataFrame(pd.DataFrame({'text':text_to_viz}))
        data, stranger_features, output_datatype = DataConversionUtils.to_spark_df(text_to_viz, self.spark,
                                                                                   self.raw_text_column)
        anno_res = self.vanilla_transformer_pipe.transform(data)
        anno_res = anno_res.collect()[0]
        if self.has_licensed_components == False:
            HTML = VizUtils.viz_OS(anno_res, self, viz_type, viz_colors, labels_to_viz, is_databricks_env,
                                   write_to_streamlit, streamlit_key,
                                   ner_col,
                                   pos_col,
                                   dep_untyped_col,
                                   dep_typed_col,

                                   )
        else:
            HTML = VizUtils.viz_HC(anno_res, self, viz_type, viz_colors, labels_to_viz, is_databricks_env,
                                   write_to_streamlit,
                                   ner_col,
                                   pos_col,
                                   dep_untyped_col,
                                   dep_typed_col,
                                   resolution_col,
                                   relation_col,
                                   assertion_col,

                                   )
        if return_html or is_databricks_env: return HTML

    def viz_streamlit(self,
                      # Base Params
                      text: Union[str, List[
                          str], pd.DataFrame, pd.Series] = "Angela Merkel from Germany and Donald Trump from America dont share many opinions",
                      model_selection: List[str] = [],
                      # SIMILARITY PARAMS
                      similarity_texts: Tuple[str, str] = ('I love NLU <3', 'I love Streamlit <3'),
                      # UI PARAMS
                      title: str = 'NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code🚀',
                      sub_title: str = 'Play with over 1000+ scalable enterprise NLP models',
                      side_info: str = None,
                      visualizers: List[str] = (
                              "dependency_tree", "ner", "similarity", "token_features", 'classification', 'manifold'),
                      show_models_info: bool = True,
                      show_model_select: bool = True,
                      show_viz_selection: bool = False,
                      show_logo: bool = True,
                      set_wide_layout_CSS: bool = True,
                      show_code_snippets: bool = False,
                      model_select_position: str = 'side',  # main or side
                      display_infos: bool = True,
                      key: str = "NLU_streamlit",
                      display_footer: bool = True,
                      num_similarity_cols: int = 2,

                      ) -> None:
        """Display Viz in streamlit"""
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.viz_streamlit_dashboard(self,
                                                         text,
                                                         model_selection,
                                                         similarity_texts,
                                                         title,
                                                         sub_title,
                                                         side_info,
                                                         visualizers,
                                                         show_models_info,
                                                         show_model_select,
                                                         show_viz_selection,
                                                         show_logo,
                                                         set_wide_layout_CSS,
                                                         show_code_snippets,
                                                         model_select_position,
                                                         display_infos,
                                                         key,
                                                         display_footer,
                                                         num_similarity_cols
                                                         )

    def viz_streamlit_token(
            self,
            text: str = 'NLU and Streamlit go together like peanutbutter and jelly',
            title: Optional[str] = "Token features",
            sub_title: Optional[str] = 'Pick from `over 1000+ models` on the left and `view the generated features`',
            show_feature_select: bool = True,
            features: Optional[List[str]] = None,
            metadata: bool = True,
            output_level: str = 'token',
            positions: bool = False,
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit",
            show_model_select=True,
            model_select_position: str = 'side',  # main or side
            show_infos: bool = True,
            show_logo: bool = True,
            show_text_input: bool = True,

    ):
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.visualize_tokens_information(self, text, title, sub_title, show_feature_select,
                                                              features, metadata, output_level, positions,
                                                              set_wide_layout_CSS, generate_code_sample, key,
                                                              show_model_select, model_select_position, show_infos,
                                                              show_logo, show_text_input)

    def viz_streamlit_classes(
            self,  # nlu component_list
            text: Union[str, list, pd.DataFrame, pd.Series, List[str]] = (
                    'I love NLU and Streamlit and sunny days!', 'I hate rainy daiys', 'CALL NOW AND WIN 1000$M'),
            output_level: Optional[str] = 'document',
            title: Optional[str] = "Text Classification",
            sub_title: Optional[
                str] = 'View predicted `classes` and `confidences` for `hundreds of text classifiers` in `over 200 languages`',
            metadata: bool = False,
            positions: bool = False,
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key: str = "NLU_streamlit",
            show_model_selector: bool = True,
            model_select_position: str = 'side',
            show_infos: bool = True,
            show_logo: bool = True,
    ) -> None:
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.visualize_classes(self, text, output_level, title, sub_title, metadata, positions,
                                                   set_wide_layout_CSS, generate_code_sample, key, show_model_selector,
                                                   model_select_position, show_infos, show_logo)

    def viz_streamlit_dep_tree(
            self,  # nlu component_list
            text: str = 'Billy likes to swim',
            title: Optional[str] = "Dependency Parse & Part-of-speech tags",
            sub_title: Optional[
                str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit",
            show_infos: bool = True,
            show_logo: bool = True,
            show_text_input: bool = True,
    ) -> None:
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.visualize_dep_tree(self, text, title, sub_title, set_wide_layout_CSS,
                                                    generate_code_sample, key, show_infos, show_logo, show_text_input, )

    def viz_streamlit_ner(
            self,  # Nlu component_list
            text: str = 'Donald Trump from America and Angela Merkel from Germany do not share many views.',
            ner_tags: Optional[List[str]] = None,
            show_label_select: bool = True,
            show_table: bool = False,
            title: Optional[str] = "Named Entities",
            sub_title: Optional[
                str] = "Recognize various `Named Entities (NER)` in text entered and filter them. You can select from over `100 languages` in the dropdown.",
            colors: Dict[str, str] = {},
            show_color_selector: bool = False,
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit",
            model_select_position: str = 'side',  # main or side
            show_model_select=True,
            show_infos: bool = True,
            show_logo: bool = True,
            show_text_input: bool = True,

    ):
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.visualize_ner(self, text, ner_tags, show_label_select, show_table, title, sub_title,
                                               colors, show_color_selector, set_wide_layout_CSS, generate_code_sample,
                                               key, model_select_position, show_model_select, show_infos, show_logo,
                                               show_text_input)

    def viz_streamlit_word_similarity(
            self,  # nlu component_list
            texts: Union[Tuple[str, str], List[str]] = (
                    "Donald Trump likes to party!", "Angela Merkel likes to party!"),
            threshold: float = 0.5,
            title: Optional[str] = "Vectors & Scalar Similarity & Vector Similarity & Embedding Visualizations  ",
            sub_tile: Optional[
                str] = "Visualize a `word-wise similarity matrix` and calculate `similarity scores` for `2 texts` and every `word embedding` loaded",
            write_raw_pandas: bool = False,
            display_embed_information: bool = True,
            similarity_matrix=True,
            show_algo_select: bool = True,
            dist_metrics: List[str] = ('cosine'),
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key: str = "NLU_streamlit",
            num_cols: int = 2,
            display_scalar_similarities: bool = False,
            display_similarity_summary: bool = False,
            model_select_position: str = 'side',
            show_infos: bool = True,
            show_logo: bool = True,

    ):
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.display_word_similarity(self, texts, threshold, title, sub_tile, write_raw_pandas,
                                                         display_embed_information, similarity_matrix, show_algo_select,
                                                         dist_metrics, set_wide_layout_CSS, generate_code_sample, key,
                                                         num_cols, display_scalar_similarities,
                                                         display_similarity_summary, model_select_position, show_infos,
                                                         show_logo, )

    def viz_streamlit_word_embed_manifold(self,
                                          default_texts: List[str] = (
                                                  "Donald Trump likes to party!", "Angela Merkel likes to party!",
                                                  'Peter HATES TO PARTTY!!!! :('),
                                          title: Optional[
                                              str] = "Lower dimensional Manifold visualization for word embeddings",
                                          sub_title: Optional[
                                              str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
                                          write_raw_pandas: bool = False,
                                          default_algos_to_apply: List[str] = ('TSNE', 'PCA',),
                                          target_dimensions: List[int] = (1, 2, 3),
                                          show_algo_select: bool = True,
                                          show_embed_select: bool = True,
                                          show_color_select: bool = True,
                                          MAX_DISPLAY_NUM: int = 100,
                                          display_embed_information: bool = True,
                                          set_wide_layout_CSS: bool = True,
                                          num_cols: int = 3,
                                          model_select_position: str = 'side',  # side or main
                                          key: str = "NLU_streamlit",
                                          additional_classifiers_for_coloring: List[str] = ['pos', 'sentiment.imdb'],
                                          generate_code_sample: bool = False,
                                          show_infos: bool = True,
                                          show_logo: bool = True,
                                          n_jobs: Optional[int] = 3,  # False
                                          ):
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.viz_streamlit_word_embed_manifold(self,
                                                                   default_texts,
                                                                   title,
                                                                   sub_title,
                                                                   write_raw_pandas,
                                                                   default_algos_to_apply,
                                                                   target_dimensions,
                                                                   show_algo_select,
                                                                   show_embed_select,
                                                                   show_color_select,
                                                                   MAX_DISPLAY_NUM,
                                                                   display_embed_information,
                                                                   set_wide_layout_CSS,
                                                                   num_cols,
                                                                   model_select_position,
                                                                   key,
                                                                   additional_classifiers_for_coloring,
                                                                   generate_code_sample,
                                                                   show_infos,
                                                                   show_logo,
                                                                   n_jobs, )

    def viz_streamlit_sentence_embed_manifold(self,
                                              default_texts: List[str] = (
                                                      "Donald Trump likes to party!", "Angela Merkel likes to party!",
                                                      'Peter HATES TO PARTTY!!!! :('),
                                              title: Optional[
                                                  str] = "Lower dimensional Manifold visualization for sentence embeddings",
                                              sub_title: Optional[
                                                  str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
                                              write_raw_pandas: bool = False,
                                              default_algos_to_apply: List[str] = ('TSNE', 'PCA',),
                                              target_dimensions: List[int] = (1, 2, 3),
                                              show_algo_select: bool = True,
                                              show_embed_select: bool = True,
                                              show_color_select: bool = True,
                                              MAX_DISPLAY_NUM: int = 100,
                                              display_embed_information: bool = True,
                                              set_wide_layout_CSS: bool = True,
                                              num_cols: int = 3,
                                              model_select_position: str = 'side',  # side or main
                                              key: str = "NLU_streamlit",
                                              additional_classifiers_for_coloring: List[str] = ['sentiment.imdb'],
                                              generate_code_sample: bool = False,
                                              show_infos: bool = True,
                                              show_logo: bool = True,
                                              n_jobs: Optional[int] = 3,  # False
                                              ):
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.viz_streamlit_sentence_embed_manifold(self,
                                                                       default_texts,
                                                                       title,
                                                                       sub_title,
                                                                       write_raw_pandas,
                                                                       default_algos_to_apply,
                                                                       target_dimensions,
                                                                       show_algo_select,
                                                                       show_embed_select,
                                                                       show_color_select,
                                                                       MAX_DISPLAY_NUM,
                                                                       display_embed_information,
                                                                       set_wide_layout_CSS,
                                                                       num_cols,
                                                                       model_select_position,
                                                                       key,
                                                                       additional_classifiers_for_coloring,
                                                                       generate_code_sample,
                                                                       show_infos,
                                                                       show_logo,
                                                                       n_jobs, )

    def viz_streamlit_entity_embed_manifold(self,
                                            default_texts: List[str] = ("Donald Trump likes to visit New York",
                                                                        "Angela Merkel likes to visit Berlin!",
                                                                        'Peter hates visiting Paris'),
                                            title: Optional[
                                                str] = "Lower dimensional Manifold visualization for Entity embeddings",
                                            sub_title: Optional[
                                                str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Entity Embeddings` to `1-D`, `2-D` and `3-D` ",
                                            default_algos_to_apply: List[str] = ("TSNE", "PCA"),
                                            target_dimensions: List[int] = (1, 2, 3),
                                            show_algo_select: bool = True,
                                            set_wide_layout_CSS: bool = True,
                                            num_cols: int = 3,
                                            model_select_position: str = 'side',  # side or main
                                            key: str = "NLU_streamlit",
                                            show_infos: bool = True,
                                            show_logo: bool = True,
                                            n_jobs: Optional[int] = 3,  # False
                                            ):
        try_import_streamlit()
        from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        StreamlitVizBlockHandler.viz_streamlit_entity_embed_manifold(self,
                                                                     default_texts,
                                                                     title,
                                                                     sub_title,
                                                                     default_algos_to_apply,
                                                                     target_dimensions,
                                                                     show_algo_select,
                                                                     set_wide_layout_CSS,
                                                                     num_cols,
                                                                     model_select_position,
                                                                     key,
                                                                     show_infos,
                                                                     show_logo,
                                                                     n_jobs)

    def __configure_light_pipe_usage__(self, data_instances, use_multi=True, force=False):
        logger.info("Configuring Light Pipeline Usage")
        if data_instances > 50 or use_multi is False:
            logger.info("Disabling light pipeline")
            if not self.is_fitted:
                self.fit()
        else:
            if not self.light_transformer_pipe or force:
                if not self.is_fitted:
                    self.fit()
                self.light_pipe_configured = True
                logger.info("Enabling light pipeline")
                self.light_transformer_pipe = LightPipeline(self.vanilla_transformer_pipe, parse_embeddings=True)

    @staticmethod
    def __get_finisher_conf(finisher: NluComponent) -> Dict[str, FinisherExtractorConfig]:
        """
        returns a dict where key=col name and value=FinisherExtractorConfig for that col for pipe and finisher.
        For finisher we need to know for each col: if its meta-field, out_put_as_arrr and if not what are sep symbols
        """

        confs = {}
        m =  finisher.model
        as_arr = m.getOutputAsArray()
        for c in m.getOutputCols():
            confs[c] = FinisherExtractorConfig(output_as_array=as_arr,
                                               is_meta_field=True if c.endswith('_metadata') else False,
                                               annotation_split_symbol=m.getAnnotationSplitSymbol(),
                                               value_split_symbol=m.getValueSplitSymbol(),
                                               source_col_name=c,
                                               )

        return confs





class RelationExtractionDL:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import RelationExtractionDLModel
        return RelationExtractionDLModel.pretrained(name = 'redl_bodypart_direction_biobert') \
                   .setInputCols(["entities", "sentence"]) \
                   .setOutputCol("relations")


    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        from sparknlp_jsl.annotator import RelationExtractionDLModel
        return RelationExtractionDLModel.pretrained(name, language,bucket) \
            .setInputCols(["entities", "sentence"]) \
            .setOutputCol("relations")

    # def get_default_trainable_model():
    #     return RelationExtractionApproach()\
    #         .setInputCols(["word_embeddings", "pos", "chunk", "unlabeled_dependency"]) \
    #         .setOutputCol("relations") \
    #         .setLabelCol('label')



from sparknlp.annotator import *

class SparkNLPLemmatizer:
    @staticmethod
    def get_default_model():
        return LemmatizerModel.pretrained() \
            .setInputCols(["token"]) \
            .setOutputCol("lemma")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return LemmatizerModel.pretrained(name, language) \
            .setInputCols(["token"]) \
            .setOutputCol("lemma")

    @staticmethod
    def get_default_trainable_model():
        return Lemmatizer \
            .setInputCols("token") \
            .setOutputCol("lemma")



from dataclasses import dataclass

from nlu.pipe.nlu_component import NluComponent
from nlu.universe.component_universes import ComponentUniverse
from nlu.universe.feature_node_ids import NLP_NODE_IDS, NLP_HC_NODE_IDS, OCR_NODE_IDS
from nlu.universe.feature_universes import NLP_FEATURES, OCR_FEATURES


### ____ Annotator Feature Resolutions ____

@dataclass
class ResolvedFeature:
    nlu_ref: str
    nlp_ref: str
    language: str
    get_pretrained: bool  # Call get_pretrained(nlp_ref, lang, bucket) or get_default() on the AnnotatorClass
    nlu_component: NluComponent  # Resolving component_to_resolve


class FeatureResolutions:
    # Map each requested Feature to a pre-defined optimal resolution, given by FeatureNode
    # Also We need Alternative Default whether licensed or not!!
    # Ideally we define nlu_ref for each of these
    # default_resolutions: Dict[JslFeature,JslAnnoId] = None
    # TODO use lang families, i.e. en.tokenize works for all Latin style languages but not Chinese, I.e. not actually multi lingual

    default_OS_resolutions = {
        NLP_FEATURES.DOCUMENT_QUESTION: ResolvedFeature('multi_document_assembler', 'multi_document_assembler', 'xx',
                                                        False,
                                                        ComponentUniverse.components[
                                                            NLP_NODE_IDS.MULTI_DOCUMENT_ASSEMBLER]),
        NLP_FEATURES.AUDIO: ResolvedFeature('audio_assembler', 'audio_assembler', 'xx', False,
                                            ComponentUniverse.components[NLP_NODE_IDS.AUDIO_ASSEMBLER]),

        NLP_FEATURES.ASSEMBLED_TABULAR_DATA: ResolvedFeature('table_assembler', 'table_assembler', 'xx', False,
                                            ComponentUniverse.components[NLP_NODE_IDS.TABLE_ASSEMBLER]),

        # NLP_FEATURES.DOCUMENT_QUESTION_CONTEXT: ResolvedFeature('multi_document_assembler', 'multi_document_assembler', 'xx', False,
        #                                        ComponentUniverse.components[NLP_NODE_IDS.MULTI_DOCUMENT_ASSEMBLER]),

        NLP_FEATURES.DOCUMENT: ResolvedFeature('document_assembler', 'document_assembler', 'xx', False,
                                               ComponentUniverse.components[NLP_NODE_IDS.DOCUMENT_ASSEMBLER]),
        NLP_FEATURES.TOKEN: ResolvedFeature('en.tokenize', 'spark_nlp_tokenizer', 'en', False,
                                            ComponentUniverse.components[NLP_NODE_IDS.TOKENIZER]),

        NLP_FEATURES.SENTENCE: ResolvedFeature('detect_sentence', 'sentence_detector_dl', 'en', False,
                                               ComponentUniverse.components[NLP_NODE_IDS.SENTENCE_DETECTOR_DL]),
        NLP_FEATURES.SENTENCE_EMBEDDINGS: ResolvedFeature('en.embed_sentence.small_bert_L2_128',
                                                          'sent_small_bert_L2_128', 'en', True,
                                                          ComponentUniverse.components[
                                                              NLP_NODE_IDS.BERT_SENTENCE_EMBEDDINGS]),
        NLP_FEATURES.WORD_EMBEDDINGS: ResolvedFeature('en.embed.bert.small_L2_128', 'small_bert_L2_128', 'en', True,
                                                      ComponentUniverse.components[NLP_NODE_IDS.BERT_EMBEDDINGS]),
        NLP_FEATURES.POS: ResolvedFeature('en.pos', 'pos_anc', 'en', True,
                                          ComponentUniverse.components[NLP_NODE_IDS.POS]),
        NLP_FEATURES.NAMED_ENTITY_IOB: ResolvedFeature('en.ner.onto.bert.cased_base', 'onto_bert_base_cased', 'en',
                                                       True,
                                                       ComponentUniverse.components[NLP_NODE_IDS.NER_DL]),

        NLP_FEATURES.NAMED_ENTITY_CONVERTED: ResolvedFeature('ner_converter', 'ner_converter', 'xx', False,
                                                             ComponentUniverse.components[NLP_NODE_IDS.NER_CONVERTER]),
        NLP_FEATURES.UNLABLED_DEPENDENCY: ResolvedFeature('en.dep.untyped', 'dependency_conllu', 'en', True,
                                                          ComponentUniverse.components[
                                                              NLP_NODE_IDS.UNTYPED_DEPENDENCY_PARSER]),
        NLP_FEATURES.LABELED_DEPENDENCY: ResolvedFeature('en.dep.typed', 'dependency_typed_conllu', 'en', True,
                                                         ComponentUniverse.components[
                                                             NLP_NODE_IDS.TYPED_DEPENDENCY_PARSER]),

        NLP_FEATURES.CHUNK: ResolvedFeature('en.chunk', 'default_chunker', 'xx', False,
                                            ComponentUniverse.components[NLP_NODE_IDS.CHUNKER]),

        NLP_FEATURES.DOCUMENT_FROM_CHUNK: ResolvedFeature(NLP_NODE_IDS.CHUNK2DOC, NLP_NODE_IDS.CHUNK2DOC, 'xx', False,
                                                          ComponentUniverse.components[NLP_NODE_IDS.CHUNK2DOC]),
        NLP_FEATURES.CHUNK_EMBEDDINGS: ResolvedFeature('en.embed_chunk', 'chunk_embeddings', 'xx', False,
                                                       ComponentUniverse.components[
                                                           NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER]),

        NLP_FEATURES.IMAGE: ResolvedFeature('image_assembler', 'image_assembler', 'xx', False,
                                            ComponentUniverse.components[
                                                NLP_NODE_IDS.IMAGE_ASSEMBLER]),

    }

    default_HC_resolutions = {
        # TODO we need ideal resolution for each lang and domain...!
        NLP_FEATURES.NAMED_ENTITY_IOB: ResolvedFeature('en.med_ner.jsl', 'ner_jsl', 'en',
                                                       True,
                                                       ComponentUniverse.components[NLP_HC_NODE_IDS.MEDICAL_NER]),

        NLP_FEATURES.NAMED_ENTITY_CONVERTED: ResolvedFeature(NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL,
                                                             NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL, 'xx', False,
                                                             ComponentUniverse.components[
                                                                 NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL]),

    }

    default_HC_train_resolutions = {

        NLP_FEATURES.NAMED_ENTITY_CONVERTED: ResolvedFeature(NLP_NODE_IDS.DOC2CHUNK, NLP_NODE_IDS.DOC2CHUNK, 'xx',
                                                             False,
                                                             ComponentUniverse.components[NLP_NODE_IDS.DOC2CHUNK]),
    }

    default_OCR_resolutions = {
        OCR_FEATURES.OCR_IMAGE: ResolvedFeature(OCR_NODE_IDS.BINARY2IMAGE, OCR_NODE_IDS.BINARY2IMAGE, 'xx', False,
                                                ComponentUniverse.components[OCR_NODE_IDS.BINARY2IMAGE]),
        OCR_FEATURES.HOCR: ResolvedFeature(OCR_NODE_IDS.IMAGE2HOCR, OCR_NODE_IDS.IMAGE2HOCR, 'xx', False,
                                           ComponentUniverse.components[OCR_NODE_IDS.IMAGE2HOCR]),
        OCR_FEATURES.IMG_SPLIT_REGIONS: ResolvedFeature(OCR_NODE_IDS.IMAGE_SPLIT_REGIONS, OCR_NODE_IDS.IMAGE_SPLIT_REGIONS, 'xx', False,
                                                        ComponentUniverse.components[OCR_NODE_IDS.IMAGE_SPLIT_REGIONS]),
        OCR_FEATURES.OCR_REGION: ResolvedFeature(OCR_NODE_IDS.IMAGE_TABLE_DETECTOR,
                                                 OCR_NODE_IDS.IMAGE_TABLE_DETECTOR, 'xx', False,
                                                 ComponentUniverse.components[OCR_NODE_IDS.IMAGE_TABLE_DETECTOR]),
        OCR_FEATURES.OCR_TABLE_CELLS: ResolvedFeature(OCR_NODE_IDS.IMAGE_TABLE_CELL_DETECTOR,
                                                      OCR_NODE_IDS.IMAGE_TABLE_CELL_DETECTOR, 'xx', False,
                                                      ComponentUniverse.components[OCR_NODE_IDS.IMAGE_TABLE_CELL_DETECTOR]),

    }
